#include <Python.h>
#include <fcntl.h>


/* Backported from Python3.3 posixmodule.c */
static PyObject *
filestore_error(void)
{
    return PyErr_SetFromErrno(PyExc_OSError);
}

static int
_parse_off_t(PyObject* arg, void* addr)
{
#if !defined(HAVE_LARGEFILE_SUPPORT)
    *((off_t*)addr) = PyLong_AsLong(arg);
#else
    *((off_t*)addr) = PyLong_AsLongLong(arg);
#endif
    if (PyErr_Occurred())
        return 0;
    return 1;
}

static PyObject *
filestore_posix_fadvise(PyObject *self, PyObject *args)
{
    off_t len, offset;
    int res, fd, advice;

    if (!PyArg_ParseTuple(args, "iO&O&i:posix_fadvise",
            &fd, _parse_off_t, &offset, _parse_off_t, &len, &advice))
        return NULL;

    Py_BEGIN_ALLOW_THREADS
    res = posix_fadvise(fd, offset, len, advice);
    Py_END_ALLOW_THREADS
    if (res != 0) {
        errno = res;
        return filestore_error();
    }
    Py_RETURN_NONE;
}


/* Our custom fallocate */
static PyObject *
filestore_fallocate(PyObject *self, PyObject *args)
{
    int res, fd, mode;
    off_t offset, len;

    if (!PyArg_ParseTuple(args, "iiO&O&:fallocate",
            &fd, &mode, _parse_off_t, &offset, _parse_off_t, &len))
        return NULL;

    Py_BEGIN_ALLOW_THREADS
    res = fallocate(fd, mode, offset, len);
    Py_END_ALLOW_THREADS
    if (res != 0) {
        errno = res;
        return filestore_error();
    }
    Py_RETURN_NONE;
}


/* Our custom fastread() */
static PyObject *
filestore_fastread(PyObject *self, PyObject *args)
{
    int fd, index, size;
    Py_ssize_t n;
    PyObject *buffer;

    off_t offset;
    int res = 0;

    if (!PyArg_ParseTuple(args, "iii:fastread", &fd, &index, &size))
        return NULL;
    if (index < 0) {
        errno = EINVAL;
        return filestore_error();
    }
    if (size < 0) {
        errno = EINVAL;
        return filestore_error();
    }
    buffer = PyBytes_FromStringAndSize((char *)NULL, size);
    if (buffer == NULL)
        return NULL;
    if (!_PyVerify_fd(fd)) {
        Py_DECREF(buffer);
        return filestore_error();
    }

    Py_BEGIN_ALLOW_THREADS
    n = read(fd, PyBytes_AS_STRING(buffer), size);
    if (n == size) {
        offset = (off_t) (index + 1) * size;
        res = posix_fadvise(fd, offset, (off_t) size, POSIX_FADV_WILLNEED);
    }
    Py_END_ALLOW_THREADS

    if (n < 0) {
        Py_DECREF(buffer);
        return filestore_error();
    }
    if (res != 0) {
        Py_DECREF(buffer);
        errno = res;
        return filestore_error();
    }
    if (n != size)
        _PyBytes_Resize(&buffer, n);
    return buffer;
}


/* module init */
static struct PyMethodDef filestore_functions[] = {
    {"fallocate", filestore_fallocate, METH_VARARGS,
     "fallocate(fd, mode, offset, len)"},
    {"posix_fadvise", filestore_posix_fadvise, METH_VARARGS,
     "posix_fadvise(fd, offset, len, advice)"},
    {"fastread", filestore_fastread, METH_VARARGS,
     "fastread(fd, leaf_index, leaf_size)"},
    {NULL, NULL, 0, NULL}
};

static struct PyModuleDef filestore = {
    PyModuleDef_HEAD_INIT,
    "_filestore",
    NULL,
    -1,
    filestore_functions
};

PyMODINIT_FUNC
PyInit__filestore(void)
{
    return PyModule_Create(&filestore);
}
