#!/usr/bin/python3

# filestore: dmedia hashing protocol and file layout
# Copyright (C) 2011 Novacut Inc
#
# This file is part of `filestore`.
#
# `filestore` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `filestore` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License along
# with `filestore`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Benchmarking utilities for in-tree tests (not installed).

Clear page cache like this:

    sync ; echo 3 | sudo tee /proc/sys/vm/drop_caches
"""

import sys
from os import path
import time
import math

import filestore
from filestore.misc import TempFileStore


BYTES10 = (
    'bytes',
    'kB',
    'MB',
    'GB',
    'TB',
    'PB',
    'EB',
    'ZB',
    'YB',
)

def bytes10(size):
    """
    Return *size* bytes to 3 significant digits in SI base-10 units.

    For example:

    >>> bytes10(1000)
    '1 kB'
    >>> bytes10(29481537)
    '29.5 MB'
    >>> bytes10(392012353)
    '392 MB'

    For additional details, see:

        https://wiki.ubuntu.com/UnitsPolicy
    """
    if size is None:
        return None
    if size < 0:
        raise ValueError('size must be >= 0; got {!r}'.format(size))
    if size == 0:
        return '0 bytes'
    if size == 1:
        return '1 byte'
    i = min(int(math.floor(math.log(size, 1000))), len(BYTES10) - 1)
    s = (size / (1000 ** i) if i > 0 else size)
    return (
        '{:.3g} {}'.format(s, BYTES10[i])
    )


class Timer(object):
    def __init__(self):
        self.start = time.time()

    def done(self):
        self.end = time.time()
        self.elapsed = self.end - self.start
        return self.elapsed

    def rate(self, total_size):
        per_sec = total_size / self.elapsed
        return '{} per second'.format(bytes10(per_sec))


if __name__ == '__main__':
    args = sys.argv[1:]
    if not args:
        raise SystemExit('ERROR: must supply source directory')
    d = path.abspath(args[0])
    if not path.isdir(d):
        raise SystemExit(
            'ERROR: source is not a directory: {!r}'.format(d)
        )
    if len(args) > 1:
        dst = filestore.FileStore(path.abspath(args[1]))
    else:
        dst = TempFileStore()

    print('Benchmarking {!r}'.format(d))
    batch = filestore.scandir(d)
    print('  {} files'.format(batch.count))
    print('  total size: {}'.format(bytes10(batch.size)))
    t = Timer()
    for (file, ch) in filestore.batch_import_iter(batch, dst):
        pass
    print('Finished in {:.3f}'.format(t.done()))
    print('Read at {}'.format(t.rate(batch.size)))
