# filestore: Dmedia hashing protocol and file layout
# Copyright (C) 2012 Novacut Inc
#
# This file is part of `filestore`.
#
# `filestore` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `filestore` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License along
# with `filestore`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Helper functions for creating a utilizing protocol test vectors.
"""

from os import path
import json
import tempfile
import shutil
from hashlib import md5
from collections import OrderedDict

from . import FileStore
from .protocols import b32dumps


TEST_VECTORS = path.join(
    path.dirname(path.abspath(__file__)), 'data', 'test-vectors.json'
)
assert path.isfile(TEST_VECTORS)


def load_test_vectors():
    return json.load(open(TEST_VECTORS, 'r'))


def build_test_leaves(leaf_size):
    return {
        'A': b'A',
        'B': b'B' * (leaf_size - 1),
        'C': b'C' * leaf_size,
    }


def build_test_vectors(protocol):
    leaves = build_test_leaves(protocol.leaf_size)

    def hash_root(letters):
        leaf_hashes = b''
        file_size = 0
        for (i, L) in enumerate(letters):
            data = leaves[L]
            leaf_hashes += protocol.hash_leaf(i, data)
            file_size += len(data)
        return protocol.hash_root(file_size, leaf_hashes)

    vectors = {
        'leaf_hashes': {},
        'root_hashes': {},
    }

    for (key, data) in leaves.items():
        vectors['leaf_hashes'][key] = [
            b32dumps(protocol.hash_leaf(i, data)) for i in range(2)
        ]

    for L in leaves:
        for key in (L, 'C' + L):
            vectors['root_hashes'][key] = hash_root(key)

    return vectors


def get_test_integers(leaf_size):
    return [
        0,
        1,
        leaf_size - 1,
        leaf_size,
        leaf_size + 1,
        2 * leaf_size - 1,
        2 * leaf_size,
    ]


def build_test_md5(leaf_size):
    integers = dict(
        (str(i), md5(str(i).encode('utf-8')).hexdigest())
        for i in get_test_integers(leaf_size)
    )
    leaves = build_test_leaves(leaf_size)
    C = leaves['C']
    files = {}
    for (key, data) in leaves.items():
        files[key] = md5(data).hexdigest()
        files['C' + key] = md5(C + data).hexdigest()
    return {
        'integers': integers,
        'files': files,
    }


class TempFileStore(FileStore):
    """
    A throw-away `FileStore` for experimenting and testing.

    When you create a `TempFileStore` instance, a random, temporary
    `FileStore.parentdir` is created for you.  For example:

    >>> fs = TempFileStore()

    Then when the `TempFileStore` instance is garbage collected, this
    directory is automatically deleted (along with any files contained
    therein).

    As with a `FileStore`, you can provide the `FileStore.id` and
    `FileStore.copies` attributes.  The `FileStore` itself ignores these
    values, but they are convenient for high-level software like Dmedia.

    For example:

    >>> fs = TempFileStore('hello', 3)
    >>> fs.id
    'hello'
    >>> fs.copies
    3

    """
    def __init__(self, _id=None, copies=0):
        parentdir = tempfile.mkdtemp(prefix='TempFileStore.')
        super().__init__(parentdir, _id, copies)
        assert self.parentdir is parentdir

    def __del__(self):
        if path.isdir(self.parentdir):
            shutil.rmtree(self.parentdir)

