# filestore: Dmedia hashing protocol and file layout
# Copyright (C) 2012 Novacut Inc
#
# This file is part of `filestore`.
#
# `filestore` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `filestore` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License along
# with `filestore`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Generic API for the Dmedia Hashing Protocol, both current and future versions.

Dmedia is a very loosely coupled distributed object store, and in order for it
to achieve its goals, two key constraints were needed:

    1. File in Dmedia are read-only (you can't modify an existing file, you can
       only insert an entire new file)

    2. Files can only be addressed by their content-hash

We feel the properties we get are well-worth these limitations, but assigning
global file IDs based on their content hash is not without peril.  For some
background on the challenges and limitations of content-based addressing, see
this excellent paper by Valerie Aurora:

    http://valerieaurora.org/monkey.html

Our take-away from the above paper is that:

    * We should assume the hashing protocol will have a limited useful
      lifetime, perhaps not longer than 10 years or so

    * We should today have in place exact plans for how a new version of the
      hashing protocol will be introduced and how multiple versions of the
      protocol can coexist side-by-side during a long transition period

In short, our migration strategy is to use the digest length as a way to select
the appropriate protocol version.  For example, say the version 1 hashing
protocol has a 280-bit digest.  And then in a decade from now when a new
protocol is needed, the new protocol will have a different digest length, say
320-bit.

This allows files hashed under different versions of the protocol to be stored
together in a FileStore.  At any given time there will be one or more valid
digest lengths, each one selecting a specific protocol version.  So if you
wanted to verify every file in a FileStore, you would iterate through them one
by one, select the protocol version based on the ID length, and then see if the
hash you compute matches the ID.

So that our reference implementation is well architected for the inevitable,
we're going to make all relevant code paths select the protocol version based
on the ID length, even when there is only one protocol available.

We want to practice now, rather than later.  This will force us to work within
this constraint, and will mean we develop design patterns that are prepared for
the future.

We might also do a full practice migration from the current protocol in use
(which we're calling version zero), to the proposed version one protocol.

Although we've never made any official commitment to the version zero protocol,
it's been around long enough that we really need at least migration
functionality, if not to support it alongside the version one protocol.
Despite not being production ready, some people have built fairly large Dmedia
libraries already.

For reference, the version zero protocol has been unchanged since November 2011:

    https://launchpad.net/filestore/trunk/11.09

In hind sight, we should have used a smaller digest size than expected for the
version zero protocol, and then planned to do a practice protocol change by at
least upping the digest size, even if no other changes were made.

In our opinion, a 240-bit digest size is probably the best choice for the
version one protocol, but unfortunately we can't use that if we want to put
Dmedia through a realistic migration scenario.  And the valuable lessons we'd
learn from such a practice scenario probably outweigh the downside not having
what we currently feel is the ideal digest size.

But even if multiple protocols are only supported in the context of unit tests,
we still feel this is a very import architecture to embrace now.
"""

from base64 import b32encode, b32decode
from _skein import skein512


# Skein personalization strings used in the Dmedia Hashing Protocol:
PERS_LEAF = b'20110430 jderose@novacut.com dmedia/leaf'
PERS_ROOT = b'20110430 jderose@novacut.com dmedia/root'

# Additional Skein personalization strings used only in the old protocol:
PERS_LEAF_INDEX = b'20110430 jderose@novacut.com dmedia/leaf-index'
PERS_FILE_SIZE = b'20110430 jderose@novacut.com dmedia/file-size'

# Provide very clear TypeError messages:
TYPE_ERROR = '{}: need a {!r}; got a {!r}: {!r}'

# Handy:
MiB = 1024**2

# For now so we don't have any issues round-tripping values through JavaScript,
# we're setting a "soft" file-size limit at 2**53 bytes (the max integer that
# fully works in JavaScript).  But this is quite big, approximately 9.01 PB for
# a single file:
MAX_FILE_SIZE = 2**53

# A real protocol would unlikely use a leaf-size this small, but a 2 MiB
# minimum means we can use a uint32_t for the leaf_index in C
# implementations and still reach the above MAX_FILE_SIZE:
MIN_LEAF_SIZE = 2 * MiB


def b32dumps(data):
    """
    Base32-encode *data*, return as a string.

    >>> b32dumps(b'bbbbb')
    'MJRGEYTC'

    """
    return b32encode(data).decode('utf-8')


def b32loads(string):
    """
    Decode base32-encoded data in *string*.

    >>> b32loads('MJRGEYTC')
    b'bbbbb'

    """
    return b32decode(string.encode('utf-8'))


class Protocol:
    """
    Standard API for current and future versions of the Dmedia Hashing Protocol.

    This class provides a standard API that abstracts away the details of a
    particular hashing protocol and allows multiple protocol versions to be
    supported simultaneously.

    The API consists of two methods:

        `Protocol.hash_leaf(leaf_index, leaf_data, challange=b'')`

        `Protocol.hash_root(file_size, leaf_hashes)`

    And four attributes:

        `Protocol.leaf_size`

        `Protocol.digest_bits`

        `Protocol.digest_bytes`

        `Protocol.digest_b32len`

    To make it easy to experiment with different *leaf_size* and *digest_bits*
    (while still using the standard hashing functions), you can simply create
    a `Protocol` instance like this:

    >>> experimental = Protocol(16 * MiB, 360)

    The `digest_bytes` and `digest_b32len` will be derived from the provided
    *digest_bits*:

    >>> experimental.digest_bytes
    45
    >>> experimental.digest_b32len
    72

    In order to implement a protocol that uses different underlying hash
    functions, you'll need to subclass `Protocol` and override the two
    low-level hashing methods:

        `Protocol._hash_leaf(leaf_index, leaf_data, challange)`

        `Protocol._hash_root(file_size, leaf_hashes)` 
    """

    def __init__(self, leaf_size, digest_bits):
        """
        Initialize a new `Protocol` instance.

        :param leaf_size: an ``int`` which is some multiple of MIN_LEAF_SIZE
            and >= MIN_LEAF_SIZE
        :param digest_bits: an ``int`` which is some multiple of 40 and >= 200
        """
        if not isinstance(leaf_size, int):
            raise TypeError(
                TYPE_ERROR.format('leaf_size', int, type(leaf_size), leaf_size)
            )
        if leaf_size < MIN_LEAF_SIZE:
            raise ValueError('leaf_size: must be >= {}; got {}'.format(
                    MIN_LEAF_SIZE, leaf_size
                )
            )
        if leaf_size % MIN_LEAF_SIZE != 0:
            raise ValueError(
                'leaf_size: must be multiple of {}; got {}'.format(
                    MIN_LEAF_SIZE, leaf_size
                )
            )
        if not isinstance(digest_bits, int):
            raise TypeError(
                TYPE_ERROR.format(
                    'digest_bits', int, type(digest_bits), digest_bits
                )
            )
        if digest_bits < 200:
            raise ValueError(
                'digest_bits: must be >= 200; got {!r}'.format(digest_bits)
            )
        if digest_bits % 40 != 0:
            raise ValueError(
                'digest_bits: must be multiple of 40; got {!r}'.format(
                    digest_bits
                )
            )
        self.leaf_size = leaf_size
        self.max_leaf_count = MAX_FILE_SIZE // leaf_size
        self.digest_bits = digest_bits
        self.digest_bytes = digest_bits // 8
        self.digest_b32len = digest_bits // 5

    def encode(self, digest):
        if not isinstance(digest, bytes):
            raise TypeError(
                TYPE_ERROR.format('digest', bytes, type(digest), digest)
            )
        if len(digest) != self.digest_bytes:
            raise ValueError('len(digest) must be {}; got {}'.format(
                    self.digest_bytes, len(digest))
            )
        return self._encode(digest)

    def _encode(self, digest):
        return b32dumps(digest)

    def decode(self, _id):
        if not isinstance(_id, str):
            raise TypeError(
                TYPE_ERROR.format('_id', str, type(_id), _id)
            )
        if len(_id) != self.digest_b32len:
            raise ValueError('len(_id) must be {}; got {}'.format(
                    self.digest_b32len, len(_id))
            )
        return self._decode(_id)

    def _decode(self, _id):
        return b32loads(_id)

    def hash_leaf(self, leaf_index, leaf_data, challenge=b''):
        """
        Hash the leaf at (zero) index *leaf_index* containing *leaf_data*.

        For example:

        >>> p = Protocol(MIN_LEAF_SIZE, 200)
        >>> b32encode(p.hash_leaf(0, b'The Leaf Data'))
        b'EFBXKNNDAMRFIIPTIL2DYZ36YYQDAX54AAOYA7WG'

        The optional *challenge* keyword argument should be a random nonce used
        to make a remote cloud storage service prove that they actually have
        the complete, exact file stored.  For example:

        >>> b32encode(p.hash_leaf(0, b'The Leaf Data', b'Random Nonce'))
        b'LBAECE4GW3UFXREWAFBMBGV2O4GISUTJ3ALZXGJW'

        :param leaf_index: an ``int`` >= 0
        :param leaf_data: a ``bytes`` instance with leaf content
        :param challenge: a ``bytes`` instance containing a random nonce to be
            used in a proof-of-storage challenge; default is ``b''``
        """
        if not isinstance(leaf_index, int):
            raise TypeError(
                TYPE_ERROR.format(
                    'leaf_index', int, type(leaf_index), leaf_index
                )
            )
        if leaf_index < 0:
            raise ValueError(
                'leaf_index: must be >= 0; got {!r}'.format(leaf_index)
            )
        if leaf_index >= self.max_leaf_count:
            raise ValueError(
                'leaf_index: must be < {!r}; got {!r}'.format(
                    self.max_leaf_count, leaf_index
                )
            )
        if not isinstance(leaf_data, bytes):
            raise TypeError(
                TYPE_ERROR.format(
                    'leaf_data', bytes, type(leaf_data), leaf_data
                )
            )
        if not (1 <= len(leaf_data) <= self.leaf_size):
            raise ValueError('Need 1 <= len(leaf_data) <= {}; got {}'.format(
                    self.leaf_size, len(leaf_data)
                )
            )
        digest = self._hash_leaf(leaf_index, leaf_data, challenge)
        assert len(digest) == self.digest_bytes
        return digest

    def hash_root(self, file_size, leaf_hashes):
        """
        Return the root-hash for a file *file_size* bytes with *leaf_hashes*.

        For example:

        >>> p = Protocol(MIN_LEAF_SIZE, 200)
        >>> p.hash_root(1776, b'DDDDDDDDDDDDDDDDDDDDDDDDD')
        'SJGXSO43OF5424G6HAEG55X2SFCZRW7L7LG42UKN'

        :param file_size: an ``int`` >= 1
        :param leaf_hashes: a ``bytes`` instance containing the concatenated
            leaf-hashes produced by `Protocol.hash_leaf()`.
        """
        if not isinstance(file_size, int):
            raise TypeError(
                TYPE_ERROR.format('file_size', int, type(file_size), file_size)
            )
        if file_size < 1:
            raise ValueError(
                'file_size: must be >= 1; got {!r}'.format(file_size)
            )
        if file_size > MAX_FILE_SIZE:
            raise ValueError(
                'file_size: must be <= {!r}; got {!r}'.format(
                    MAX_FILE_SIZE, file_size
                )
            )
        if not isinstance(leaf_hashes, bytes):
            raise TypeError(TYPE_ERROR.format(
                    'leaf_hashes', bytes, type(leaf_hashes), leaf_hashes
                )
            )
        if len(leaf_hashes) == 0:
            raise ValueError('leaf_hashes cannot be empty')
        if len(leaf_hashes) % self.digest_bytes != 0:
            raise ValueError(
                'len(leaf_hashes) is {}, not multiple of {}'.format(
                    len(leaf_hashes), self.digest_bytes
                )
            )
        n = len(leaf_hashes) // self.digest_bytes
        low = (n - 1) * self.leaf_size + 1
        high = n * self.leaf_size
        if not (low <= file_size <= high):
            raise ValueError(
                'Need {} <= file_size <= {}; got {}'.format(
                    low, high, file_size
                )
            )
        digest = self._hash_root(file_size, leaf_hashes)
        return self.encode(digest)

    def _hash_leaf(self, leaf_index, leaf_data, challenge):
        """
        Protocol version 1 leaf-hashing implementation.

        Subclasses can override this to use different hash functions,
        configurations, etc.
        """
        assert leaf_index >= 0
        assert 1 <= len(leaf_data) <= self.leaf_size
        return skein512(leaf_data,
            digest_bits=self.digest_bits,
            pers=PERS_LEAF,
            key=str(leaf_index).encode('utf-8'),
            nonce=challenge,
        ).digest()

    def _hash_root(self, file_size, leaf_hashes):
        """
        Protocol version 1 root-hashing implementation.

        Subclasses can override this to use different hash functions,
        configurations, etc.
        """
        assert file_size >= 1
        assert len(leaf_hashes) > 0
        assert len(leaf_hashes) % self.digest_bytes == 0
        return skein512(leaf_hashes,
            digest_bits=self.digest_bits,
            pers=PERS_ROOT,
            key=str(file_size).encode('utf-8'),
        ).digest()


VERSION1 = Protocol(8 * MiB, 280)


class OldProtocol(Protocol):
    """
    The unofficial version zero protocol.

    Although no formal support commitments were ever made for this version zero
    protocol, it has been used enough in the wild that we should really
    continue to support it and provide a nice migration to the version one
    protocol.
    """

    def _hash_leaf_index(self, leaf_index):
        assert leaf_index >= 0
        return skein512(str(leaf_index).encode('utf-8'),
            digest_bits=self.digest_bits,
            pers=PERS_LEAF_INDEX,
        ).digest()

    def _hash_leaf(self, leaf_index, leaf_data, challenge):
        assert leaf_index >= 0
        assert 1 <= len(leaf_data) <= self.leaf_size
        skein = skein512(self._hash_leaf_index(leaf_index),
            digest_bits=self.digest_bits,
            pers=PERS_LEAF,
            nonce=challenge,
        )
        skein.update(leaf_data)
        return skein.digest()

    def _hash_file_size(self, file_size):
        assert file_size >= 1
        return skein512(str(file_size).encode('utf-8'),
            digest_bits=self.digest_bits,
            pers=PERS_FILE_SIZE,
        ).digest()

    def _hash_root(self, file_size, leaf_hashes):
        assert file_size >= 1
        assert len(leaf_hashes) > 0
        assert len(leaf_hashes) % self.digest_bytes == 0
        skein = skein512(self._hash_file_size(file_size),
            digest_bits=self.digest_bits,
            pers=PERS_ROOT,
        )
        skein.update(leaf_hashes)
        return skein.digest()


VERSION0 = OldProtocol(8 * MiB, 240)

PROTOCOLS = (VERSION0, VERSION1)
