# filestore: dmedia hashing protocol and file layout
# Copyright (C) 2011 Novacut Inc
#
# This file is part of `filestore`.
#
# `filestore` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `filestore` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License along
# with `filestore`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Unit tests for `filestore` module.
"""

from unittest import TestCase
import os
from os import path
import io
import stat
from base64 import b32encode, b32decode
from subprocess import check_call
import tempfile
import shutil
from random import SystemRandom

from skein import skein512

from filestore.protocols import PROTOCOLS, VERSION0, VERSION1
from filestore.protocols import b32dumps, b32loads
from filestore import misc
import filestore


random = SystemRandom()
TYPE_ERROR = '{}: need a {!r}; got a {!r}: {!r}'

B32NAMES = []
for a in filestore.B32ALPHABET:
    for b in filestore.B32ALPHABET:
        B32NAMES.append(a + b)
B32NAMES.sort()
B32NAMES = tuple(B32NAMES)


# Sample file:
LEAVES = (
    b'a' * filestore.LEAF_SIZE,
    b'b' * filestore.LEAF_SIZE,
    b'c' * (filestore.LEAF_SIZE // 2),
)
FILE_SIZE = (filestore.LEAF_SIZE * 2) + (filestore.LEAF_SIZE // 2)
LEAF_HASHES = b''.join([
    filestore.hash_leaf(0, LEAVES[0]),
    filestore.hash_leaf(1, LEAVES[1]),
    filestore.hash_leaf(2, LEAVES[2]),
])
ID = filestore.hash_root(FILE_SIZE, LEAF_HASHES)
CH = filestore.ContentHash(ID, FILE_SIZE, LEAF_HASHES)
assert CH.id == 'DMJGE4OTWZVKSX426GHE46GZCGICMQOVWNGRVB7E665Y2RAM', CH.id


def write_sample_file(fp):
    if isinstance(fp, str):
        fp = open(fp, 'wb')
    for leaf in LEAVES:
        fp.write(leaf)
    fp.close()


def tobytes(n):
    return str(n).encode('utf-8')


def random_id(id_bytes=filestore.DIGEST_BYTES):
    return b32encode(os.urandom(id_bytes)).decode('ascii')


class TempDir(object):
    def __init__(self, prefix='unit-tests.'):
        self.dir = tempfile.mkdtemp(prefix=prefix)

    def join(self, *parts):
        return path.join(self.dir, *parts)

    def rmtree(self):
        if self.dir is not None:
            check_call(['/bin/chmod', '-R', '+w', self.dir])
            shutil.rmtree(self.dir)
            self.dir = None

    def makedirs(self, *parts):
        d = self.join(*parts)
        if not path.exists(d):
            os.makedirs(d)
        return d

    def touch(self, *parts):
        d = self.makedirs(*parts[:-1])
        f = self.join(*parts)
        assert not path.exists(f)
        open(f, 'wb').close()
        assert path.isfile(f) and not path.islink(f)
        return f

    def write(self, content, *parts):
        d = self.makedirs(*parts[:-1])
        f = self.join(*parts)
        assert not path.exists(f)
        open(f, 'wb').write(content)
        assert path.isfile(f) and not path.islink(f)
        return f

    def copy(self, src, *parts):
        self.makedirs(*parts[:-1])
        dst = self.join(*parts)
        assert not path.exists(dst)
        shutil.copy2(src, dst)
        assert path.isfile(dst) and not path.islink(dst)
        return dst

    def __del__(self):
        self.rmtree()


class DummyQueue:
    def __init__(self):
        self.items = []
  
    def put(self, item):
        self.items.append(item)


class Callback:
    def __init__(self):
        self.calls = []

    def __call__(self, *args):
        self.calls.append(args)


class TestSmartQueue(TestCase):
    def test_get(self):
        q = filestore.SmartQueue()

        # Test with some common types that will go through the queue:
        leaf = filestore.Leaf(0, b'the leaf data')
        q.put(leaf)
        q.put(None)
        q.put(filestore.EndBatch)
        self.assertIs(q.get(), leaf)
        self.assertIsNone(q.get())
        self.assertIs(q.get(), filestore.EndBatch)

        # Test with some exceptions
        ex1 = ValueError('oh no')
        ex2 = Exception('ouch')
        ex3 = filestore.FileNotFound('the id', 'the store')
        q.put(ex1)
        q.put(ex2)
        q.put(ex3)
        with self.assertRaises(ValueError) as cm:
            item = q.get()
        self.assertIs(cm.exception, ex1)
        with self.assertRaises(Exception) as cm:
            item = q.get()
        self.assertIs(cm.exception, ex2)
        with self.assertRaises(filestore.FileNotFound) as cm:
            item = q.get()
        self.assertIs(cm.exception, ex3)

        # Test that only Exception *instances* are raised
        q.put(filestore.FileNotFound)
        q.put(TypeError)
        self.assertIs(q.get(), filestore.FileNotFound)
        self.assertIs(q.get(), TypeError)


class TestIDError(TestCase):
    def test_init(self):
        e= filestore.IDError('foo')
        self.assertEqual(e.id, 'foo')
        self.assertEqual(str(e), "invalid file ID: 'foo'")
        
        
class TestLeafHashesError(TestCase):
    def test_init(self):
        e = filestore.LeafHashesError(b'Hello')
        self.assertEqual(e.leaf_hashes, b'Hello')
        self.assertEqual(str(e), 'invalid leaf-hashes')


class TestRootHashError(TestCase):
    def test_init(self):
        e = filestore.RootHashError('foo', 18, b'leaf hashes', 'bar')
        self.assertEqual(e.id, 'foo')
        self.assertEqual(e.file_size, 18)
        self.assertEqual(e.leaf_hashes, b'leaf hashes')
        self.assertEqual(e.bad_id, 'bar')
        self.assertEqual(str(e), "'foo'")


class TestNestedParent(TestCase):
    def test_init(self):
        e = filestore.NestedParent('/home/joe/.dmedia/sue', '.dmedia')
        self.assertEqual(e.parentdir, '/home/joe/.dmedia/sue')
        self.assertEqual(e.dotname, '.dmedia')
        self.assertEqual(str(e), "'/home/joe/.dmedia/sue' contains '.dmedia'")


class TestPathError(TestCase):
    def test_init(self):
        e = filestore.PathError('/foo/../bar', '/bar')
        self.assertEqual(e.untrusted, '/foo/../bar')
        self.assertEqual(e.abspath, '/bar')
        self.assertEqual(str(e), "'/foo/../bar' resolves to '/bar'")


class TestPathTraversal(TestCase):
    def test_init(self):
        class DummyFileStore:
            basedir = '/home/user/.dmedia'
        
        e = filestore.PathTraversal(
            '/home/user/.dmeida/files/../../.ssh',
            '/home/user/.ssh',
            DummyFileStore
        )
        self.assertEqual(e.untrusted, '/home/user/.dmeida/files/../../.ssh')
        self.assertEqual(e.abspath, '/home/user/.ssh')
        self.assertIs(e.store, DummyFileStore)
        self.assertEqual(
            str(e),
            "'/home/user/.ssh' outside of '/home/user/.dmedia'"
        )


class TestCorruptFile(TestCase):
    def test_init(self):
        e = filestore.CorruptFile('the id', 'the store', foo='bar')
        self.assertEqual(e.id, 'the id')
        self.assertEqual(e.store, 'the store')
        self.assertEqual(e.kw, {'foo': 'bar'})
        self.assertEqual(
            str(e),
            "'the id' is corrupted in 'the store'"
        )

        e = filestore.CorruptFile('the id', 'the store')
        self.assertEqual(e.kw, {})


class TestFileNotFound(TestCase):
    def test_init(self):
        e = filestore.FileNotFound('the id', 'the FileStore')
        self.assertEqual(e.id, 'the id')
        self.assertEqual(e.store, 'the FileStore')
        self.assertEqual(str(e), "'the id' not in 'the FileStore'")


class TestFunctions(TestCase):
    def test_constants(self):
        """
        Sanity checks on various constants.
        """
        self.assertEqual(len(filestore.B32ALPHABET), 32)

        self.assertEqual(filestore.B32NAMES, B32NAMES)
        self.assertEqual(len(filestore.B32NAMES), 1024)

        self.assertTrue(filestore.DOTNAME.startswith('.'))

        # Must be multiple of 1 MiB
        self.assertEqual(filestore.LEAF_SIZE % (2**20), 0)

        # Must be multiple of 40 bits for base32-encoding
        self.assertEqual(filestore.DIGEST_BITS % 40, 0)

        # And a multiple of 8 bits, of course
        self.assertEqual(filestore.DIGEST_BITS % 8, 0)

        # Test the stuff derived from DIGEST_BITS
        self.assertIsInstance(filestore.DIGEST_BYTES, int)
        self.assertEqual(filestore.DIGEST_BYTES * 8, filestore.DIGEST_BITS)
        self.assertEqual(filestore.DIGEST_BYTES, 30)

        self.assertIsInstance(filestore.DIGEST_B32LEN, int)
        self.assertEqual(filestore.DIGEST_B32LEN * 5, filestore.DIGEST_BITS)
        self.assertEqual(filestore.DIGEST_B32LEN, 48)

    def test_hash_leaf_index(self):
        f = filestore.hash_leaf_index

        # Test with wrong leaf_index type:
        with self.assertRaises(TypeError) as cm:
            pers = f(0.0)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('leaf_index', int, float, 0.0)
        )
        with self.assertRaises(TypeError) as cm:
            pers = f(1.5)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('leaf_index', int, float, 1.5)
        )

        # Test with wrong leaf_index value
        with self.assertRaises(ValueError) as cm:
            pers = f(-1)
        self.assertEqual(
            str(cm.exception),
            'leaf_index: must be >= 0; got -1'
        )
        with self.assertRaises(ValueError) as cm:
            pers = f(-17)
        self.assertEqual(
            str(cm.exception),
            'leaf_index: must be >= 0; got -17'
        )

        accum = set()
        for i in range(1000):
            expected = skein512(str(i).encode('utf-8'),
                digest_bits=240,
                pers=b'20110430 jderose@novacut.com dmedia/leaf-index',
            ).digest()
            self.assertEqual(f(i), expected)
            self.assertNotIn(expected, accum)
            accum.add(expected)

        self.assertEqual(
            b32encode(f(0)),
            b'23NC2GGK42HGXSJKSCUEIEHNUSHTAPLYDDKQ7P7ZZZ7VWYDH'
        )

    def test_hash_leaf(self):
        f = filestore.hash_leaf

        # Test with wrong leaf_index type:
        with self.assertRaises(TypeError) as cm:
            pers = f(0.0)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('leaf_index', int, float, 0.0)
        )
        with self.assertRaises(TypeError) as cm:
            pers = f(1.5)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('leaf_index', int, float, 1.5)
        )

        # Test with wrong leaf_index value
        with self.assertRaises(ValueError) as cm:
            pers = f(-1)
        self.assertEqual(
            str(cm.exception),
            'leaf_index: must be >= 0; got -1'
        )
        with self.assertRaises(ValueError) as cm:
            pers = f(-17)
        self.assertEqual(
            str(cm.exception),
            'leaf_index: must be >= 0; got -17'
        )

        # Test with wrong leaf_data type
        leaf_data = 'a' * 31415
        with self.assertRaises(TypeError) as cm:
            digest = f(2, leaf_data)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('leaf_data', bytes, str, leaf_data)
        )

        # Test with wrong leaf_data length
        toosmall = b''
        with self.assertRaises(ValueError) as cm:
            digest = f(2, toosmall)
        self.assertEqual(
            str(cm.exception),
            'Need 1 <= len(leaf_data) <= 8388608; got 0'
        )
        toobig = b'a' * 8388609
        with self.assertRaises(ValueError) as cm:
            digest = f(2, toobig)
        self.assertEqual(
            str(cm.exception),
            'Need 1 <= len(leaf_data) <= 8388608; got 8388609'
        )

        # Test with good values:
        lh = f(2)
        self.assertEqual(lh.block_bits, 512)
        self.assertEqual(lh.digest_bits, 240)
        self.assertEqual(
            lh.digest(),
            skein512(filestore.hash_leaf_index(2),
                digest_bits=240,
                pers=filestore.PERS_LEAF,
            ).digest()
        )

        content = b'N'
        digest = f(2, content)
        self.assertEqual(
            digest,
            skein512(filestore.hash_leaf_index(2) + content,
                digest_bits=240,
                pers=filestore.PERS_LEAF,
            ).digest()
        )

        content = b'N' * filestore.LEAF_SIZE
        digest = f(2, content)
        self.assertEqual(
            digest,
            skein512(filestore.hash_leaf_index(2) + content,
                digest_bits=240,
                pers=filestore.PERS_LEAF,
            ).digest()
        )

        # A 25k value sanity check on our crytographic claim that the
        # leaf_index is tied to the leaf_hash:
        count = 25 * 1000
        leaf_data = os.urandom(16)
        accum = set(
            filestore.hash_leaf(i, leaf_data)
            for i in range(count)
        )
        self.assertEqual(len(accum), count)

        # A 25k random value sanity check on our crytographic claim that the
        # leaf_data is tied to the leaf_hash:
        accum = set(
            filestore.hash_leaf(21, os.urandom(16))
            for i in range(count)
        )
        self.assertEqual(len(accum), count)

    def test_hash_file_size(self):
        f = filestore.hash_file_size

        # Test with wrong file_size type:
        with self.assertRaises(TypeError) as cm:
            f(1.0)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('file_size', int, float, 1.0)
        )
        with self.assertRaises(TypeError) as cm:
            f(1.5)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('file_size', int, float, 1.5)
        )

        # Test with bad file_size value
        with self.assertRaises(ValueError) as cm:
            f(0)
        self.assertEqual(
            str(cm.exception),
            'file_size: must be >= 1; got 0'
        )
        with self.assertRaises(ValueError) as cm:
            f(-1)
        self.assertEqual(
            str(cm.exception),
            'file_size: must be >= 1; got -1'
        )

        accum = set()
        for i in range(1, 100):
            expected = skein512(str(i).encode('utf-8'),
                digest_bits=240,
                pers=b'20110430 jderose@novacut.com dmedia/file-size',
            ).digest()
            self.assertEqual(f(i), expected)
            self.assertNotIn(expected, accum)
            accum.add(expected)

        self.assertEqual(
            b32encode(f(1)),
            b'ITRK32RKWWCM4MUILND5UJ7SWSTVGCZPD4KC2HISQHMLW2WL'
        )

    def test_hash_root(self):
        f = filestore.hash_root

        # Test with wrong file_size type:
        with self.assertRaises(TypeError) as cm:
            pers = f(1.0, None)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('file_size', int, float, 1.0)
        )
        with self.assertRaises(TypeError) as cm:
            pers = f(1.5, None)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('file_size', int, float, 1.5)
        )

        # Test with bad file_size value
        with self.assertRaises(ValueError) as cm:
            pers = f(0, None)
        self.assertEqual(
            str(cm.exception),
            'file_size: must be >= 1; got 0'
        )
        with self.assertRaises(ValueError) as cm:
            pers = f(-1, None)
        self.assertEqual(
            str(cm.exception),
            'file_size: must be >= 1; got -1'
        )

        # Test with wrong leaf_hashes type:
        leaf_hashes = 'A' * 30
        with self.assertRaises(TypeError) as cm:
            rh = filestore.hash_root(17, leaf_hashes)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('leaf_hashes', bytes, str, leaf_hashes)
        )

        # Test with empty leaf_hashes
        with self.assertRaises(ValueError) as cm:
            rh = filestore.hash_root(17, b'')
        self.assertEqual(str(cm.exception), 'leaf_hashes cannot be empty')

        # Test when len(leaf_hashes) is not a multiple of DIGEST_BYTES
        with self.assertRaises(ValueError) as cm:
            rh = filestore.hash_root(17, b'A' * 45)
        self.assertEqual(
            str(cm.exception),
            'len(leaf_hashes) is 45, not multiple of 30'
        )

        # Test low and high bounds
        with self.assertRaises(ValueError) as cm:
            f(filestore.LEAF_SIZE + 1, b'a' * filestore.DIGEST_BYTES)
        self.assertEqual(
            str(cm.exception),
            'Need 1 <= file_size <= 8388608; got 8388609'
        )
        with self.assertRaises(ValueError) as cm:
            f(filestore.LEAF_SIZE, b'ab' * filestore.DIGEST_BYTES)
        self.assertEqual(
            str(cm.exception),
            'Need 8388609 <= file_size <= 16777216; got 8388608'
        )
        with self.assertRaises(ValueError) as cm:
            f(2 * filestore.LEAF_SIZE + 1, b'ab' * filestore.DIGEST_BYTES)
        self.assertEqual(
            str(cm.exception),
            'Need 8388609 <= file_size <= 16777216; got 16777217'
        )

        # Test with good file_size and leaf_hashes
        leaf_hashes = b'D' * filestore.DIGEST_BYTES
        self.assertEqual(
            filestore.hash_root(1, leaf_hashes),
            '4AZOU4R7M6JKJJRQMVX42YB7ULRUCS6FZGJNZCDVOATXYPML'
        )
        self.assertEqual(
            filestore.hash_root(filestore.LEAF_SIZE, leaf_hashes),
            'RXBEFCKXWKYNPXZOR234QHYI475L2AF7C4AOQUG7EG7UJBDJ'
        )

        # A 25k value sanity check on our crytographic claim that the
        # file_size is tied to the root_hash:
        count = 25 * 1000
        leaf_hashes = os.urandom(filestore.DIGEST_BYTES)
        accum = set(
            filestore.hash_root(size, leaf_hashes)
            for size in range(1, count + 1)
        )
        self.assertEqual(len(accum), count)

        # A 25k random value sanity check on our crytographic claim that the
        # leaf_hashes are tied to the root_hash:
        accum = set(
            filestore.hash_root(314159, os.urandom(filestore.DIGEST_BYTES))
            for i in range(count)
        )
        self.assertEqual(len(accum), count)

    def test_check_leaf_hashes(self):
        # Test with wrong type:
        leaf_hashes = 'a' * 60
        with self.assertRaises(TypeError) as cm:
            filestore.check_leaf_hashes(leaf_hashes)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('leaf_hashes', bytes, str, leaf_hashes)
        )

        # Test with empty leaf_hashes:
        leaf_hashes = b''
        with self.assertRaises(filestore.LeafHashesError) as cm:
            filestore.check_leaf_hashes(leaf_hashes)
        self.assertIs(cm.exception.leaf_hashes, leaf_hashes)

        # Test with wrong length:
        leaf_hashes = b'a' * 61
        with self.assertRaises(filestore.LeafHashesError) as cm:
            filestore.check_leaf_hashes(leaf_hashes)
        self.assertIs(cm.exception.leaf_hashes, leaf_hashes)

        # Test with good values:
        good30 = os.urandom(30)
        self.assertIs(filestore.check_leaf_hashes(good30), good30)
        self.assertIs(
            filestore.check_leaf_hashes(good30, protocol=VERSION0),
            good30
        )
        good35 = os.urandom(35)
        self.assertIs(
            filestore.check_leaf_hashes(good35, protocol=VERSION1),
            good35
        )

        # Test with wrong protocol version:
        with self.assertRaises(filestore.LeafHashesError) as cm:
            filestore.check_leaf_hashes(good30, protocol=VERSION1)
        self.assertIs(cm.exception.leaf_hashes, good30)
        with self.assertRaises(filestore.LeafHashesError) as cm:
            filestore.check_leaf_hashes(good35, protocol=VERSION0)
        self.assertIs(cm.exception.leaf_hashes, good35)

        # Future proofing:
        for protocol in PROTOCOLS:
            good = os.urandom(protocol.digest_bytes)
            self.assertIs(filestore.check_leaf_hashes(good, protocol), good)

            good = os.urandom(18 * protocol.digest_bytes)
            self.assertIs(filestore.check_leaf_hashes(good, protocol), good)

            bad = os.urandom(protocol.digest_bytes + 1)
            with self.assertRaises(filestore.LeafHashesError) as cm:
                filestore.check_leaf_hashes(bad, protocol)
            self.assertIs(cm.exception.leaf_hashes, bad)

            bad = os.urandom(18 * protocol.digest_bytes - 1)
            with self.assertRaises(filestore.LeafHashesError) as cm:
                filestore.check_leaf_hashes(bad, protocol)
            self.assertIs(cm.exception.leaf_hashes, bad)

    def test_check_root_hash(self):
        f = filestore.check_root_hash

        self.assertEqual(f(ID, FILE_SIZE, LEAF_HASHES), CH)
        self.assertEqual(
            f(ID, FILE_SIZE, LEAF_HASHES, unpack=True),
            filestore.ContentHash(ID, FILE_SIZE, 
                tuple(filestore.iter_leaf_hashes(LEAF_HASHES))
            )
        )

        with self.assertRaises(filestore.RootHashError) as cm:
            f(ID, FILE_SIZE + 1, LEAF_HASHES)
        self.assertEqual(cm.exception.id, ID)
        self.assertEqual(cm.exception.file_size, FILE_SIZE + 1)
        self.assertEqual(cm.exception.leaf_hashes, LEAF_HASHES)
        self.assertEqual(cm.exception.bad_id,
            'U5FK5XRT33ZJTCYIO3WJB7YTTGESXCBKZEW35J7FIQI7UN7S'
        )

        with self.assertRaises(filestore.RootHashError) as cm:
            f(ID, FILE_SIZE - 1, LEAF_HASHES)
        self.assertEqual(cm.exception.id, ID)
        self.assertEqual(cm.exception.file_size, FILE_SIZE - 1)
        self.assertEqual(cm.exception.leaf_hashes, LEAF_HASHES)
        self.assertEqual(cm.exception.bad_id,
            'OYESBWEZ4Y2AGSLMNZB4ZF75A2VG7NXVB4R25SSMRGXLN4CR'
        )

    def test_check_root_hash2(self):
        for protocol in PROTOCOLS:
            file_size = protocol.leaf_size + 1
            leaf0_hash = os.urandom(protocol.digest_bytes)
            leaf1_hash = os.urandom(protocol.digest_bytes)
            leaf_hashes = leaf0_hash + leaf1_hash
            _id = protocol.hash_root(file_size, leaf_hashes)

            # Everything is correct:
            ch = filestore.check_root_hash(_id, file_size, leaf_hashes)
            self.assertIsInstance(ch, filestore.ContentHash)
            self.assertEqual(ch.id, _id)
            self.assertEqual(ch.file_size, file_size)
            self.assertEqual(ch.leaf_hashes, leaf_hashes)
            self.assertEqual(ch, (_id, file_size, leaf_hashes))

            # Everything is correct, unpack=True
            ch = filestore.check_root_hash(_id, file_size, leaf_hashes,
                unpack=True
            )
            self.assertIsInstance(ch, filestore.ContentHash)
            self.assertEqual(ch.id, _id)
            self.assertEqual(ch.file_size, file_size)
            self.assertEqual(ch.leaf_hashes, (leaf0_hash, leaf1_hash))
            self.assertEqual(ch, (_id, file_size, (leaf0_hash, leaf1_hash)))

            # Wrong ID:
            bad_id = random_id(protocol.digest_bytes)
            with self.assertRaises(filestore.RootHashError) as cm:
                filestore.check_root_hash(bad_id, file_size, leaf_hashes)
            self.assertEqual(cm.exception.id, bad_id)
            self.assertEqual(cm.exception.file_size, file_size)
            self.assertEqual(cm.exception.leaf_hashes, leaf_hashes)
            self.assertEqual(cm.exception.bad_id, _id)

            # Wrong file_size:
            with self.assertRaises(filestore.RootHashError) as cm:
                filestore.check_root_hash(_id, file_size + 1, leaf_hashes)
            self.assertEqual(cm.exception.id, _id)
            self.assertEqual(cm.exception.file_size, file_size + 1)
            self.assertEqual(cm.exception.leaf_hashes, leaf_hashes)
            self.assertEqual(cm.exception.bad_id,
                protocol.hash_root(file_size + 1, leaf_hashes)
            )

            # Wrong leaf_hashes
            bad_leaf_hashes = leaf0_hash + os.urandom(protocol.digest_bytes)
            with self.assertRaises(filestore.RootHashError) as cm:
                filestore.check_root_hash(_id, file_size, bad_leaf_hashes)
            self.assertEqual(cm.exception.id, _id)
            self.assertEqual(cm.exception.file_size, file_size)
            self.assertEqual(cm.exception.leaf_hashes, bad_leaf_hashes)
            self.assertEqual(cm.exception.bad_id,
                protocol.hash_root(file_size, bad_leaf_hashes)
            )

    def test_enumerate_leaf_hashes(self):
        f = filestore.enumerate_leaf_hashes
        self.assertEqual(
            list(f(b'a' * 30)),
            [(0, b'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaa')]
        )
        leaf_hashes = (b'X' * 30) + (b'Y' * 30) + (b'Z' * 30)
        self.assertEqual(
            list(f(leaf_hashes)),
            [
                (0, b'XXXXXXXXXXXXXXXXXXXXXXXXXXXXXX'),
                (1, b'YYYYYYYYYYYYYYYYYYYYYYYYYYYYYY'),
                (2, b'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ'),
            ]
        )

    def test_iter_leaf_hashes(self):
        f = filestore.iter_leaf_hashes
        self.assertEqual(
            list(f(b'a' * 30)),
            [b'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaa']
        )
        leaf_hashes = (b'X' * 30) + (b'Y' * 30) + (b'Z' * 30)
        self.assertEqual(
            list(f(leaf_hashes)),
            [
                b'XXXXXXXXXXXXXXXXXXXXXXXXXXXXXX',
                b'YYYYYYYYYYYYYYYYYYYYYYYYYYYYYY',
                b'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
            ]
        )

    def test_get_leaf_hash(self):
        self.assertEqual(
            filestore.get_leaf_hash(LEAF_HASHES, 0),
            filestore.hash_leaf(0, LEAVES[0])
        )
        self.assertEqual(
            filestore.get_leaf_hash(LEAF_HASHES, 1),
            filestore.hash_leaf(1, LEAVES[1])
        )
        self.assertEqual(
            filestore.get_leaf_hash(LEAF_HASHES, 2),
            filestore.hash_leaf(2, LEAVES[2])
        )
        self.assertEqual(
            filestore.get_leaf_hash(LEAF_HASHES, 3),
            b''
        )
        self.assertEqual(
            filestore.get_leaf_hash(LEAF_HASHES, 4),
            b''
        )

    def test_missing_leaves(self):
        f = filestore.missing_leaves
        tmp = TempDir()
        chunks = (
            b'Y' * filestore.LEAF_SIZE,
            b'E' * filestore.LEAF_SIZE,
            b'S',
        )
        leaf_hashes = b''.join(
            filestore.hash_leaf(i, chunk) for (i, chunk) in enumerate(chunks)
        )
        src = tmp.touch('test.mov')

        # Test with empty file
        fp = open(src, 'rb')
        self.assertEqual(
            list(f(fp, leaf_hashes)),
            list(filestore.enumerate_leaf_hashes(leaf_hashes))
        )

        # Test with correct file
        fp = open(src, 'wb')
        for chunk in chunks:
            fp.write(chunk)
        fp = open(src, 'rb')
        self.assertEqual(
            list(f(fp, leaf_hashes)),
            [],
        )

        # Test with just the last 1-byte leaf correct:
        fp = open(src, 'wb')
        fp.write(b'N' * filestore.LEAF_SIZE)
        fp.write(b'O' * filestore.LEAF_SIZE)
        fp.write(b'S')
        fp = open(src, 'rb')
        self.assertEqual(
            list(f(fp, leaf_hashes)),
            [
                (0, filestore.hash_leaf(0, chunks[0])),
                (1, filestore.hash_leaf(1, chunks[1])),
            ]
        )

    def test_check_path(self):
        f = filestore.check_path

        # Test with relative path:
        with self.assertRaises(filestore.PathError) as cm:
            f('foo/bar')
        self.assertEqual(cm.exception.untrusted, 'foo/bar')
        self.assertEqual(cm.exception.abspath, path.abspath('foo/bar'))

        # Test with path traversal:
        with self.assertRaises(filestore.PathError) as cm:
            f('/foo/bar/../../root')
        self.assertEqual(cm.exception.untrusted, '/foo/bar/../../root')
        self.assertEqual(cm.exception.abspath, '/root')

        # Test with normalized absolute path:
        self.assertEqual(f('/home/jderose/.dmedia'), '/home/jderose/.dmedia')

    def test_check_id(self):
        # Test with wrong type
        bad = random_id(30).encode('utf-8')
        with self.assertRaises(TypeError) as cm:
            filestore.check_id(bad)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('_id', str, bytes, bad)
        )

        # Test with wrong length:
        for i in range(321):
            value = 'N' * i
            if len(value) in (48, 56):
                self.assertIs(filestore.check_id(value), value)
                continue
            with self.assertRaises(filestore.IDError) as cm:
                filestore.check_id(value)
            self.assertIs(cm.exception.id, value)

        # Test with 48 and 56 character:
        id48 = random_id(30)
        self.assertIs(filestore.check_id(id48), id48)
        id56 = random_id(35)
        self.assertIs(filestore.check_id(id56), id56)

        # Test case sensitivity:
        bad48 = id48.lower()
        with self.assertRaises(filestore.IDError) as cm:
            filestore.check_id(bad48)
        self.assertIs(cm.exception.id, bad48)
        bad56 = id56.lower()
        with self.assertRaises(filestore.IDError) as cm:
            filestore.check_id(bad56)
        self.assertIs(cm.exception.id, bad56)

    def test_iter_files(self):
        """
        Test `iter_files()` and `scandir()`.
        """
        tmp = TempDir()
        self.assertEqual(list(filestore.iter_files(tmp.dir)), [])
        self.assertEqual(
            filestore.scandir(tmp.dir),
            filestore.Batch(tuple(), 0, 0)
        )

        good = tmp.makedirs('good')
        bad = tmp.touch('bad')
        nope = tmp.join('nope')

        # Test when basedir dose not exist:
        with self.assertRaises(OSError) as cm:
            list(filestore.iter_files(nope))
        self.assertEqual(cm.exception.errno, 2)
        with self.assertRaises(OSError) as cm:
            filestore.scandir(nope)
        self.assertEqual(cm.exception.errno, 2)

        # Test when basedir is a file rather than directory:
        with self.assertRaises(OSError) as cm:
            list(filestore.iter_files(bad))
        self.assertEqual(cm.exception.errno, 20)
        with self.assertRaises(OSError) as cm:
            filestore.scandir(bad)
        self.assertEqual(cm.exception.errno, 20)

        self.assertEqual(list(filestore.iter_files(good)), [])

        b = tmp.makedirs('good', 'b')
        c = tmp.write(b'n' * 18, 'good', 'c')
        a = tmp.write(b'n' * 314, 'good', 'a')
        d = tmp.write(b'n' * 42, 'good', 'd')

        b_a = tmp.write(b'n' * 32, 'good', 'b', 'a')
        b_b = tmp.write(b'n' * 64, 'good', 'b', 'b')
        b_c = tmp.write(b'n' * 8, 'good', 'b', 'c')
        b_d = tmp.write(b'n' * 16, 'good', 'b', 'd')

        # All symlinks should be ignored:
        bb = tmp.join('good', 'bb')
        os.symlink(b, bb)
        self.assertTrue(path.isdir(bb))
        self.assertEqual(os.readlink(bb), b)

        os.symlink(a, tmp.join('good', 'aa'))
        os.symlink(c, tmp.join('good', 'cc'))
        os.symlink(d, tmp.join('good', 'dd'))

        files = (
            filestore.File(a, 314, path.getmtime(a)),
            filestore.File(c, 18, path.getmtime(c)),
            filestore.File(d, 42, path.getmtime(d)),

            filestore.File(b_a, 32, path.getmtime(b_a)),
            filestore.File(b_b, 64, path.getmtime(b_b)),
            filestore.File(b_c, 8, path.getmtime(b_c)),
            filestore.File(b_d, 16, path.getmtime(b_d)),
        )
        total = sum(f.size for f in files)

        self.assertEqual(
            tuple(filestore.iter_files(good)),
            files,
        )
        self.assertEqual(
            filestore.scandir(good),
            filestore.Batch(files, total, 7)
        )

    def test_reader(self):
        tmp = TempDir()
        src = tmp.join('foo.mov')
        write_sample_file(src)
        src_fp = open(src, 'rb')

        q = DummyQueue()
        self.assertIsNone(filestore.reader(src_fp, q))
        self.assertEqual(len(q.items), len(LEAVES) + 1)
        for (i, data) in enumerate(LEAVES):
            leaf = q.items[i]
            self.assertIsInstance(leaf, filestore.Leaf)
            self.assertEqual(leaf, filestore.Leaf(i, data))
        self.assertIsNone(q.items[-1])
        self.assertFalse(src_fp.closed)

        # Test that src_fp.seek(0) is called:
        q = DummyQueue()
        self.assertIsNone(filestore.reader(src_fp, q))
        self.assertEqual(len(q.items), len(LEAVES) + 1)
        for (i, data) in enumerate(LEAVES):
            leaf = q.items[i]
            self.assertIsInstance(leaf, filestore.Leaf)
            self.assertEqual(leaf, filestore.Leaf(i, data))
        self.assertIsNone(q.items[-1])
        self.assertFalse(src_fp.closed)

        # Test with a file opened in r+b
        src_fp = open(src, 'r+b')
        q = DummyQueue()
        self.assertIsNone(filestore.reader(src_fp, q))
        self.assertEqual(len(q.items), len(LEAVES) + 1)
        for (i, data) in enumerate(LEAVES):
            leaf = q.items[i]
            self.assertIsInstance(leaf, filestore.Leaf)
            self.assertEqual(leaf, filestore.Leaf(i, data))
        self.assertIsNone(q.items[-1])
        self.assertFalse(src_fp.closed)

        # Test a closed file
        src_fp = open(src, 'rb')
        src_fp.close()
        q = DummyQueue()
        self.assertIsNone(filestore.reader(src_fp, q))
        self.assertEqual(len(q.items), 1)
        e = q.items[0]
        self.assertIsInstance(e, ValueError)
        self.assertEqual(str(e), 'seek of closed file')

        # Test with file open in 'wb' mode
        src_fp = open(src, 'wb')
        q = DummyQueue()
        self.assertIsNone(filestore.reader(src_fp, q))
        self.assertEqual(len(q.items), 1)
        e = q.items[0]
        self.assertIsInstance(e, TypeError)
        self.assertEqual(
            str(e),
            TYPE_ERROR.format('src_fp', filestore.IO_READABLE, type(src_fp), src_fp)
        )

        # Test with an empty file
        src = tmp.touch('empty.mov')
        src_fp = open(src, 'rb')
        q = DummyQueue()
        self.assertIsNone(filestore.reader(src_fp, q))
        self.assertEqual(q.items, [None])
        self.assertFalse(src_fp.closed)

    def test_batch_reader(self):
        tmp = TempDir()
        # Note that batch_reader() shouldn't pay attention to File.size
        # nor File.mtime, nor should it pay attention to Batch.size nor
        # Batch.count.
        src = filestore.File(tmp.join('foo.mov'), 0, 0)
        write_sample_file(src.name)
        empty = filestore.File(tmp.touch('zero.mov'), 0, 0)

        # Test with an empty batch
        batch = filestore.Batch(tuple(), 0, 0)
        q = DummyQueue()
        self.assertIsNone(filestore.batch_reader(batch, q))
        self.assertEqual(q.items, [filestore.EndBatch])

        # Test a batch with 1 empty file
        batch = filestore.Batch((empty,), 0, 0)
        q = DummyQueue()
        self.assertIsNone(filestore.batch_reader(batch, q))
        self.assertEqual(
            q.items,
            [
                empty,
                filestore.EndFile,
                filestore.EndBatch,
            ]
        )

        # Test a batch with 1 non-empty file
        batch = filestore.Batch((src,), 0, 0)
        q = DummyQueue()
        self.assertIsNone(filestore.batch_reader(batch, q))
        self.assertEqual(
            q.items,
            [
                src,
                filestore.Leaf(0, LEAVES[0]),
                filestore.Leaf(1, LEAVES[1]),
                filestore.Leaf(2, LEAVES[2]),
                filestore.EndFile,
                filestore.EndBatch,
            ]
        )

        # Test a batch with empty followed by non-empty
        batch = filestore.Batch((empty, src), 0, 0)
        q = DummyQueue()
        self.assertIsNone(filestore.batch_reader(batch, q))
        self.assertEqual(
            q.items,
            [
                empty,
                filestore.EndFile,
                src,
                filestore.Leaf(0, LEAVES[0]),
                filestore.Leaf(1, LEAVES[1]),
                filestore.Leaf(2, LEAVES[2]),
                filestore.EndFile,
                filestore.EndBatch,
            ]
        )

        # Test a batch with non-empty followed by empty
        batch = filestore.Batch((src, empty), 0, 0)
        q = DummyQueue()
        self.assertIsNone(filestore.batch_reader(batch, q))
        self.assertEqual(
            q.items,
            [
                src,
                filestore.Leaf(0, LEAVES[0]),
                filestore.Leaf(1, LEAVES[1]),
                filestore.Leaf(2, LEAVES[2]),
                filestore.EndFile,
                empty,
                filestore.EndFile,
                filestore.EndBatch,
            ]
        )

        # Test 3 empties in a row:
        one = filestore.File(tmp.touch('one.mov'), 0, 0)
        two = filestore.File(tmp.touch('two.mov'), 0, 0)
        three = filestore.File(tmp.touch('three.mov'), 0, 0)
        batch = filestore.Batch((one, two, three), 0, 0)
        q = DummyQueue()
        self.assertIsNone(filestore.batch_reader(batch, q))
        self.assertEqual(
            q.items,
            [
                one,
                filestore.EndFile,
                two,
                filestore.EndFile,
                three,
                filestore.EndFile,
                filestore.EndBatch,
            ]
        )

        # Test that exceptions are caught and put into q:
        batch = filestore.Batch('hello', 0, 0)
        q = DummyQueue()
        self.assertIsNone(filestore.batch_reader(batch, q))
        self.assertEqual(len(q.items), 2)
        self.assertEqual(q.items[0], 'h')
        e = q.items[1]
        self.assertIsInstance(e, AttributeError)
        self.assertEqual(str(e), "'str' object has no attribute 'name'")

    def test_reader_iter(self):
        tmp = TempDir()
        src = tmp.write(b'hello', 'some', 'file.txt')
        allowed = (io.BufferedReader, io.BufferedRandom)

        # Wrong src_fp type:
        with self.assertRaises(TypeError) as cm:
            for leaf in filestore.reader_iter(17):
                pass
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('src_fp', allowed, int, 17)
        )

        # File opened in wrong mode:
        src_fp = open(src, 'r')
        with self.assertRaises(TypeError) as cm:
            for leaf in filestore.reader_iter(src_fp):
                pass
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('src_fp', allowed, type(src_fp), src_fp)
        )

        # Test that exception gets raised
        src = tmp.join('my.mov')
        write_sample_file(src)
        src_fp = open(src, 'rb')
        src_fp.close()
        with self.assertRaises(ValueError) as cm:
            for leaf in filestore.reader_iter(src_fp):
                pass
        self.assertEqual(str(cm.exception), 'seek of closed file')

    def test_batch_import_iter(self):
        batch = filestore.Batch(tuple(), 0, 0)
        tmp1 = TempDir()
        fs1 = filestore.FileStore(tmp1.dir)
        tmp2 = TempDir()
        fs2 = filestore.FileStore(tmp2.dir)

        # Test with wrong batch type
        b = (tuple(), 0, 0)
        with self.assertRaises(TypeError) as cm:
            list(filestore.batch_import_iter(b, fs1, fs2))
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('batch', filestore.Batch, tuple, b)
        )

        # Test with no filestores
        with self.assertRaises(TypeError) as cm:
            list(filestore.batch_import_iter(batch))
        self.assertEqual(
            str(cm.exception),
            'batch_import_iter() takes at least one destination FileStore'
        )
        
        # Test with non-callable callback
        with self.assertRaises(TypeError) as cm:
            list(filestore.batch_import_iter(batch, fs1, fs2, callback='hello'))
        self.assertEqual(
            str(cm.exception),
            "callback not callable: 'hello'"
        )

        # Test with wrong filestore type
        with self.assertRaises(TypeError) as cm:
            list(filestore.batch_import_iter(batch, 17, fs2))
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('filestores[0]', filestore.FileStore, int, 17)
        )
        with self.assertRaises(TypeError) as cm:
            list(filestore.batch_import_iter(batch, fs1, 18))
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('filestores[1]', filestore.FileStore, int, 18)
        )

        # Test with empty batch:
        self.assertEqual(list(filestore.batch_import_iter(batch, fs1, fs2)), [])
        self.assertEqual(list(fs1), [])
        self.assertEqual(list(fs2), [])

        # Test a non-empty batch:
        tmp = TempDir()
        src = filestore.File(tmp.join('src.mov'), FILE_SIZE, 0)
        write_sample_file(src.name)
        empty1 = filestore.File(tmp.touch('empty1.mov'), 0, 0)
        empty2 = filestore.File(tmp.touch('empty2.mov'), 0, 0)
        batch = filestore.Batch((empty1, src, empty2), FILE_SIZE, 0)
        self.assertEqual(
            list(filestore.batch_import_iter(batch, fs1, fs2)),
            [
                (empty1, None),
                (src, CH),
                (empty2, None),
            ]
        )
        self.assertEqual([s.id for s in fs1], [ID])
        self.assertEqual([s.id for s in fs2], [ID])

        # Test that a permisson error propigates properly
        os.chmod(src.name, 0)
        with self.assertRaises(IOError) as cm:
            for (file, ch) in filestore.batch_import_iter(batch, fs1, fs2):
                pass
        self.assertEqual(
            str(cm.exception),
            '[Errno 13] Permission denied: {!r}'.format(src.name)
        )

        # Test with callback and empty batch
        cb = Callback()
        batch = filestore.Batch(tuple(), 0, 0)
        tmp1 = TempDir()
        fs1 = filestore.FileStore(tmp1.dir)
        tmp2 = TempDir()
        fs2 = filestore.FileStore(tmp2.dir)
        self.assertEqual(
            list(filestore.batch_import_iter(batch, fs1, fs2, callback=cb)),
            []
        )
        self.assertEqual(list(fs1), [])
        self.assertEqual(list(fs2), [])
        self.assertEqual(cb.calls, [])

        # Test with callback and non-empty batch
        tmp = TempDir()
        src = filestore.File(tmp.join('src.mov'), FILE_SIZE, 0)
        write_sample_file(src.name)
        empty1 = filestore.File(tmp.touch('empty1.mov'), 0, 0)
        empty2 = filestore.File(tmp.touch('empty2.mov'), 0, 0)

        file_size = filestore.LEAF_SIZE * 33
        big = filestore.File(tmp.join('big.mov'), file_size, 0)
        data = b'a' * filestore.LEAF_SIZE
        fp = open(big.name, 'wb')
        for i in range(33):
            fp.write(data)
        fp.close()
        fp = open(big.name, 'rb')
        ch = filestore.hash_fp(fp)

        batch = filestore.Batch((big, empty1, src, empty2), 0, 0)
        self.assertEqual(
            list(filestore.batch_import_iter(batch, fs1, fs2, callback=cb)),
            [
                (big, ch),
                (empty1, None),
                (src, CH),
                (empty2, None),
            ]
        )
        self.assertEqual(set(s.id for s in fs1), set([ID, ch.id]))
        self.assertEqual(set(s.id for s in fs2), set([ID, ch.id]))
        self.assertEqual(
            cb.calls,
            [
                (0, filestore.LEAF_SIZE * 8),  # 64 MiB done
                (0, filestore.LEAF_SIZE * 16),  # 128 MiB done
                (0, filestore.LEAF_SIZE * 24),  # 192 MiB done
                (0, filestore.LEAF_SIZE * 32),  # 256 MiB done
                (1, file_size),
                (2, file_size),
                (3, file_size + 20971520),
                (4, file_size + 20971520)
            ]
        )

    def test_hash_fp(self):
        tmp = TempDir()
        src = tmp.join('foo.mov')
        dst = tmp.join('bar.mov')
        write_sample_file(src)

        src_fp = open(src, 'rb')
        ch = filestore.hash_fp(src_fp)
        self.assertIsInstance(ch, filestore.ContentHash)
        self.assertEqual(ch.id,
            'DMJGE4OTWZVKSX426GHE46GZCGICMQOVWNGRVB7E665Y2RAM'
        )
        self.assertEqual(ch.file_size, 20971520)
        self.assertFalse(path.exists(dst))

        src_fp = open(src, 'rb')
        dst_fp = open(dst, 'wb')
        ch = filestore.hash_fp(src_fp, dst_fp)
        self.assertIsInstance(ch, filestore.ContentHash)
        self.assertEqual(ch.id,
            'DMJGE4OTWZVKSX426GHE46GZCGICMQOVWNGRVB7E665Y2RAM'
        )
        self.assertEqual(ch.file_size, 20971520)
        self.assertTrue(path.isfile(dst))

        src_fp = open(dst, 'rb')
        self.assertEqual(filestore.hash_fp(src_fp), ch)

    def test_hash_fp2(self):
        obj = misc.load_test_vectors()
        for proto in PROTOCOLS:
            tmp = TempDir()
            leaves = misc.build_test_leaves(proto.leaf_size)
            vectors = obj[str(proto.digest_b32len)]
            for (key, value) in vectors['root_hashes'].items():
                name = tmp.join(key)
                assert not path.exists(name)
                fp = open(name, 'wb')
                for L in key:
                    fp.write(leaves[L])
                fp.close()
                fp = open(name, 'rb')
                ch = filestore.hash_fp(fp, protocol=proto)
                self.assertIsInstance(ch, filestore.ContentHash)
                self.assertEqual(ch.id, value)
                self.assertEqual(ch.file_size, path.getsize(name))
                self.assertEqual(ch.leaf_hashes,
                    b''.join(
                        b32loads(vectors['leaf_hashes'][L][i])
                        for (i, L) in enumerate(key)
                    )
                )

    def test_ensuredir(self):
        f = filestore.ensuredir
        tmp = TempDir()

        # Test that os.makedirs() is not used:
        d = tmp.join('foo', 'bar')
        with self.assertRaises(OSError) as cm:
            f(d)
        self.assertEqual(
            str(cm.exception), 
            '[Errno 2] No such file or directory: {!r}'.format(d)
        )
        self.assertFalse(path.exists(d))
        self.assertFalse(path.exists(tmp.join('foo')))

        # Test when dir does not exist
        d = tmp.join('foo')
        self.assertFalse(path.exists(d))
        self.assertIs(f(d), True)
        self.assertTrue(path.isdir(d))

        # Test when dir exists:
        self.assertIs(f(d), False)
        self.assertTrue(path.isdir(d))

        # Test when dir exists and is a file:
        d = tmp.touch('bar')
        with self.assertRaises(ValueError) as cm:
            f(d)
        self.assertEqual(str(cm.exception), 'not a directory: {!r}'.format(d))

        # Test when dir exists and is a symlink to a dir:
        target = tmp.makedirs('target')
        link = tmp.join('link')
        os.symlink(target, link)
        self.assertTrue(path.isdir(link))
        self.assertTrue(path.islink(link))
        with self.assertRaises(ValueError) as cm:
            f(link)
        self.assertEqual(
            str(cm.exception),
            '{!r} is symlink to {!r}'.format(link, target)
        )

    def test_statvfs(self):
        tmp = TempDir()
        s1 = os.statvfs(tmp.dir)
        s2 = filestore.statvfs(tmp.dir)
        self.assertIsInstance(s2, filestore.StatVFS)
        # FIXME: For some reason the StatVFS.readonly attribute is failing
        # on Raring, but only for i386:
        return
        size = s1.f_frsize * s1.f_blocks
        avail = s1.f_frsize * s1.f_bavail
        free = s1.f_frsize * s1.f_bfree
        self.assertEqual(s2,
            filestore.StatVFS(size, size - free, avail, False, s1.f_frsize)
        )


class TestHasher(TestCase):
    def test_init(self):
        h = filestore.Hasher()
        self.assertEqual(h.file_size, 0)
        self.assertEqual(h.leaf_index, 0)
        self.assertIsInstance(h.array, bytearray)
        self.assertEqual(h.array, b'')
        self.assertFalse(h.closed)

    def test_hash_leaf(self):
        # Test with bad leaf type
        h = filestore.Hasher()
        with self.assertRaises(TypeError) as cm:
            h.hash_leaf(b'nope')
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('leaf', filestore.Leaf, bytes, b'nope')
        )

        # We'll use these below
        leaf0 = filestore.Leaf(0, LEAVES[0])
        leaf1 = filestore.Leaf(1, LEAVES[1])
        leaf2 = filestore.Leaf(2, LEAVES[2])

        # Test when closed
        h = filestore.Hasher()
        h.closed = True
        with self.assertRaises(Exception) as cm:
            h.hash_leaf(leaf0)
        self.assertEqual(
            str(cm.exception),
            'Cannot call Hasher.hash_leaf() when Hasher.closed'
        )

        # Test when leaf_index is wrong
        h = filestore.Hasher()
        h.leaf_index = 1
        with self.assertRaises(Exception) as cm:
            h.hash_leaf(leaf0)
        self.assertEqual(
            str(cm.exception),
            'Expected leaf.index 1, got 0'
        )
 
        # Test when it's all good
        h = filestore.Hasher()

        self.assertEqual(h.hash_leaf(leaf0), LEAF_HASHES[0:30])
        self.assertEqual(h.leaf_index, 1)
        self.assertEqual(h.file_size, filestore.LEAF_SIZE)
        self.assertFalse(h.closed)

        self.assertEqual(h.hash_leaf(leaf1), LEAF_HASHES[30:60])
        self.assertEqual(h.leaf_index, 2)
        self.assertEqual(h.file_size, filestore.LEAF_SIZE * 2)
        self.assertFalse(h.closed)

        self.assertEqual(h.hash_leaf(leaf2), LEAF_HASHES[60:90])
        self.assertEqual(h.leaf_index, 3)
        self.assertEqual(h.file_size, sum(len(l) for l in LEAVES))
        self.assertTrue(h.closed)

    def test_content_hash(self):
        leaf0 = filestore.Leaf(0, LEAVES[0])
        leaf1 = filestore.Leaf(1, LEAVES[1])
        leaf2 = filestore.Leaf(2, LEAVES[2])

        h = filestore.Hasher()
        for leaf in (leaf0, leaf1, leaf2):
            h.hash_leaf(leaf)
        self.assertEqual(h.content_hash(), CH)

        # Test that Hasher.content_hash() sets closed = True
        h = filestore.Hasher()
        h.hash_leaf(leaf0)
        self.assertFalse(h.closed)
        h.content_hash()
        self.assertTrue(h.closed)


class TestFileStore(TestCase):
    def test_init(self):

        # Test with relative path:
        with self.assertRaises(filestore.PathError) as cm:
            fs = filestore.FileStore('foo/bar')
        self.assertEqual(cm.exception.untrusted, 'foo/bar')
        self.assertEqual(cm.exception.abspath, path.abspath('foo/bar'))
        self.assertEqual(
            str(cm.exception),
            '{!r} resolves to {!r}'.format('foo/bar', path.abspath('foo/bar'))
        )

        # Test with path traversal:
        with self.assertRaises(filestore.PathError) as cm:
            fs = filestore.FileStore('/foo/bar/../../root')
        self.assertEqual(cm.exception.untrusted, '/foo/bar/../../root')
        self.assertEqual(cm.exception.abspath, '/root')
        self.assertEqual(
            str(cm.exception),
            '{!r} resolves to {!r}'.format('/foo/bar/../../root', '/root')
        )

        # Test with a nested parentdir:
        tmp = TempDir()
        parentdir = tmp.makedirs('.dmedia', 'foo')
        with self.assertRaises(filestore.NestedParent) as cm:
            fs = filestore.FileStore(parentdir)
        self.assertEqual(cm.exception.parentdir, parentdir)
        self.assertEqual(cm.exception.dotname, '.dmedia')
        self.assertEqual(
            str(cm.exception),
            '{!r} contains {!r}'.format(parentdir, '.dmedia')
        )

        # Test when parentdir doesn't exists:
        tmp = TempDir()
        parentdir = tmp.join('nope')
        with self.assertRaises(ValueError) as cm:
            fs = filestore.FileStore(parentdir)
        self.assertEqual(
            str(cm.exception),
            'FileStore.parentdir not a directory: {!r}'.format(parentdir)
        )

        # Test when parentdir is a file:
        tmp = TempDir()
        parentdir = tmp.touch('somefile')
        with self.assertRaises(ValueError) as cm:
            fs = filestore.FileStore(parentdir)
        self.assertEqual(
            str(cm.exception),
            'FileStore.parentdir not a directory: {!r}'.format(parentdir)
        )

        # Test when basedir does not exist
        tmp = TempDir()
        basedir = tmp.join('.dmedia')
        self.assertFalse(path.exists(basedir))
        fs = filestore.FileStore(tmp.dir)
        self.assertEqual(fs.parentdir, tmp.dir)
        self.assertEqual(fs.basedir, basedir)
        self.assertTrue(path.isdir(basedir))
        self.assertEqual(fs.tmp, path.join(basedir, 'tmp'))
        self.assertTrue(path.isdir(fs.tmp))
        self.assertIsNone(fs.id)
        self.assertEqual(fs.copies, 0)

        # Test when _id and copies are supplied
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir, 'foo', 1)
        self.assertEqual(fs.id, 'foo')
        self.assertEqual(fs.copies, 1)

        # Test when basedir exists and is a directory
        tmp = TempDir()
        basedir = tmp.makedirs('.dmedia')
        fs = filestore.FileStore(tmp.dir)
        self.assertEqual(fs.parentdir, tmp.dir)
        self.assertEqual(fs.basedir, basedir)
        self.assertTrue(path.isdir(basedir))
        self.assertEqual(fs.tmp, path.join(basedir, 'tmp'))
        self.assertFalse(path.isdir(fs.tmp))

        # Test when basedir exists and is a file
        tmp = TempDir()
        basedir = tmp.touch('.dmedia')
        with self.assertRaises(ValueError) as cm:
            fs = filestore.FileStore(tmp.dir)
        self.assertEqual(
            str(cm.exception),
            'not a directory: {!r}'.format(basedir)
        )

        # Test when basedir exists and is a symlink to a dir
        tmp = TempDir()
        d = tmp.makedirs('foo')
        basedir = tmp.join('.dmedia')
        os.symlink(d, basedir)
        with self.assertRaises(ValueError) as cm:
            fs = filestore.FileStore(tmp.dir)
        self.assertEqual(
            str(cm.exception),
            '{!r} is symlink to {!r}'.format(basedir, d)
        )

    def test_repr(self):
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)
        self.assertEqual(repr(fs), 'FileStore({!r})'.format(tmp.dir))

    def test_iter(self):
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)

        # Should ignore files with wrong ID length:
        short = tuple(random_id(25) for i in range(50))
        for _id in short:
            f = fs.join('files', _id[:2], _id[2:])
            assert not path.exists(f)
            open(f, 'wb').write(b'Novacut')
            os.chmod(f, stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH)
            self.assertEqual(path.getsize(f), 7)
        self.assertEqual(list(fs), [])
        long = tuple(random_id(40) for i in range(50))
        for _id in long:
            f = fs.join('files', _id[:2], _id[2:])
            assert not path.exists(f)
            open(f, 'wb').write(b'Novacut')
            os.chmod(f, stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH)
            self.assertEqual(path.getsize(f), 7)
        self.assertEqual(list(fs), [])

        # Should ignore files invalid b32 letters:
        for i in range(50):
            _id = random_id(25) + '1ABCDEFG'  # 1 is not in B32ALPHABET
            assert len(_id) == filestore.DIGEST_B32LEN
            f = fs.join('files', _id[:2], _id[2:])
            assert not path.exists(f)
            open(f, 'wb').write(b'Novacut')
            os.chmod(f, stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH)
            self.assertEqual(path.getsize(f), 7)
        self.assertEqual(list(fs), [])

        # Should ignore empty files:
        for i in range(50):
            _id = random_id()
            f = fs.join('files', _id[:2], _id[2:])
            assert not path.exists(f)
            open(f, 'wb').close()
            os.chmod(f, stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH)
            self.assertEqual(path.getsize(f), 0)
            assert path.isfile(fs.path(_id))
        self.assertEqual(list(fs), [])

        # Should ignore directories
        for i in range(50):
            _id = random_id()
            d = fs.join('files', _id[:2], _id[2:])
            assert not path.exists(d)
            os.mkdir(d)
            os.chmod(d, stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH)
            assert path.isdir(fs.path(_id))
        self.assertEqual(list(fs), [])

        # Now add valid files in (48 character IDs)
        stats = []
        for i in range(2000):
            _id = random_id(30)
            size = i + 1
            f = fs.path(_id)
            assert not path.exists(f)
            open(f, 'wb').write(b'N' * size)
            os.chmod(f, stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH)
            self.assertEqual(path.getsize(f), size)
            st = filestore.Stat(_id, f, size, path.getmtime(f))
            self.assertEqual(fs.stat(_id), st)
            stats.append(st)
        self.assertNotEqual(list(fs), stats)  # Sorting!
        stats.sort(key=lambda s: s.id)
        self.assertEqual(list(fs), stats)

        # Add more valid files in (56 character IDs)
        for i in range(1000):
            _id = random_id(35)
            size = i + 1
            f = fs.path(_id)
            assert not path.exists(f)
            open(f, 'wb').write(b'N' * size)
            os.chmod(f, stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH)
            self.assertEqual(path.getsize(f), size)
            st = filestore.Stat(_id, f, size, path.getmtime(f))
            self.assertEqual(fs.stat(_id), st)
            stats.append(st)
        self.assertNotEqual(list(fs), stats)  # Sorting!
        stats.sort(key=lambda s: s.id)
        self.assertEqual(list(fs), stats)

        # Should ignore symlinks, even if to valid files
        # This makes sure os.lstat() is being used rather than os.stat()  
        for i in range(50):
            _id = random_id()
            link = fs.path(_id)
            assert not path.exists(link)
            file = fs.path(stats[i].id)
            assert path.isfile(file) and not path.islink(file)
            os.symlink(file, link)
            assert path.isfile(fs.path(_id))
            assert path.islink(fs.path(_id))
        self.assertEqual(list(fs), stats)

    def check_subdirs(self, basedir):
        self.assertEqual(
            sorted(os.listdir(basedir)),
            ['README.txt', 'corrupt', 'files', 'partial', 'tmp']
        )

        f = path.join(basedir, 'README.txt')
        self.assertFalse(path.islink(f))
        self.assertEqual(open(f, 'rt').read(), filestore.README)

        self.assertEqual(
            sorted(os.listdir(path.join(basedir, 'files'))),
            list(B32NAMES)
        )
        for name in ['corrupt', 'partial', 'tmp']:
            d = path.join(basedir, name)
            self.assertEqual(os.listdir(d), [])

        for name in ['corrupt', 'files', 'partial', 'tmp']:
            d = path.join(basedir, name)
            self.assertTrue(path.isdir(d))
            self.assertFalse(path.islink(d))
        for name in B32NAMES:
            d = path.join(basedir, 'files', name)
            self.assertTrue(path.isdir(d))
            self.assertFalse(path.islink(d))
            self.assertEqual(os.listdir(d), [])

    def test_init_dirs(self):
        tmp = TempDir()

        # Create basedir dir so init_dirs() doesn't get called
        basedir = tmp.makedirs('.dmedia')
        fs = filestore.FileStore(tmp.dir)
        self.assertEqual(os.listdir(basedir), [])

        # Test when no subdirs exist:
        self.assertIsNone(fs.init_dirs())
        self.check_subdirs(basedir)

        # Test when all subdirs exist:
        self.assertIsNone(fs.init_dirs())
        self.check_subdirs(basedir)

        # Test when some subdirs exist:
        os.rmdir(path.join(basedir, 'tmp'))
        for (i, name) in enumerate(B32NAMES):
            if i % 3 == 0:
                d = path.join(basedir, 'files', name)
        self.assertIsNone(fs.init_dirs())
        self.check_subdirs(basedir)

    def test_statvfs(self):
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)
        s1 = os.statvfs(tmp.dir)
        s2 = fs.statvfs()
        self.assertIsInstance(s2, filestore.StatVFS)
        size = s1.f_frsize * s1.f_blocks
        avail = s1.f_frsize * s1.f_bavail
        free = s1.f_frsize * s1.f_bfree
        self.assertEqual(s2,
            filestore.StatVFS(size, size - free, avail, False, s1.f_frsize)
        )

    def test_catch_traversal(self):
        tmp = TempDir()
        parentdir = tmp.makedirs('foo')
        basedir = tmp.join('foo', '.dmedia')
        fs = filestore.FileStore(parentdir)

        bad = tmp.join('foo', '.dmedia2', 'stuff')
        with self.assertRaises(filestore.PathTraversal) as cm:
            fs.catch_traversal(bad)
        self.assertEqual(cm.exception.untrusted, bad)
        self.assertEqual(cm.exception.abspath, bad)
        self.assertIs(cm.exception.store, fs)

        bad = tmp.join('foo', '.dmedia', '..', '.dmedia2')
        assert '..' in bad
        with self.assertRaises(filestore.PathTraversal) as cm:
            fs.catch_traversal(bad)
        self.assertEqual(cm.exception.untrusted, bad)
        self.assertEqual(cm.exception.abspath, tmp.join('foo', '.dmedia2'))
        self.assertIs(cm.exception.store, fs)

        good = tmp.join('foo', '.dmedia', 'stuff')
        self.assertEqual(fs.catch_traversal(good), good)

    def test_join(self):
        tmp = TempDir()
        parentdir = tmp.makedirs('foo')
        fs = filestore.FileStore(parentdir)

        # Test with an absolute path in parts:
        with self.assertRaises(filestore.PathTraversal) as cm:
            fs.join('files', '/root')
        self.assertEqual(cm.exception.untrusted, '/root')
        self.assertEqual(cm.exception.abspath, '/root')
        self.assertIs(cm.exception.store, fs)

        # Test with some .. climbers:
        with self.assertRaises(filestore.PathTraversal) as cm:
            fs.join('NW', '..', '..', '.ssh')
        self.assertEqual(
            cm.exception.untrusted,
            tmp.join('foo', '.dmedia', 'NW', '..', '..', '.ssh')
        )
        self.assertEqual(cm.exception.abspath, tmp.join('foo', '.ssh'))
        self.assertIs(cm.exception.store, fs)

        # Test for former security issue!  See:
        # https://bugs.launchpad.net/dmedia/+bug/708663
        with self.assertRaises(filestore.PathTraversal) as cm:
            fs.join('..', '.dmedia2', 'stuff')
        self.assertEqual(
            cm.exception.untrusted,
            tmp.join('foo', '.dmedia', '..', '.dmedia2', 'stuff')
        )
        self.assertEqual(
            cm.exception.abspath,
            tmp.join('foo', '.dmedia2', 'stuff')
        )
        self.assertIs(cm.exception.store, fs)

        # Test with some correct parts:
        self.assertEqual(
            fs.join('NW', 'BNVXVK5DQGIOW7MYR4K3KA5K22W7NW'),
            tmp.join('foo', '.dmedia', 'NW', 'BNVXVK5DQGIOW7MYR4K3KA5K22W7NW')
        )

    def test_path(self):
        tmp = TempDir()
        parentdir = tmp.makedirs('foo')
        fs = filestore.FileStore(parentdir)

        _id = random_id()
        self.assertEqual(
            fs.path(_id),
            tmp.join('foo', '.dmedia', 'files', _id[:2], _id[2:])
        )

        # Test to make sure hashes are getting checked with check_id():
        bad = 'NWBNVXVK5..GIOW7MYR4K3KA5K22W7NW'
        with self.assertRaises(filestore.IDError) as cm:
            fs.path(bad)
        self.assertIs(cm.exception.id, bad)

    def test_partial_path(self):
        tmp = TempDir()
        parentdir = tmp.makedirs('foo')
        fs = filestore.FileStore(parentdir)

        _id = random_id()
        self.assertEqual(
            fs.partial_path(_id),
            tmp.join('foo', '.dmedia', 'partial', _id)
        )

        # Test to make sure hashes are getting checked with check_id():
        bad = 'NWBNVXVK5..GIOW7MYR4K3KA5K22W7NW'
        with self.assertRaises(filestore.IDError) as cm:
            fs.partial_path(bad)
        self.assertIs(cm.exception.id, bad)

    def test_corrupt_path(self):
        tmp = TempDir()
        parentdir = tmp.makedirs('foo')
        fs = filestore.FileStore(parentdir)

        _id = random_id()
        self.assertEqual(
            fs.corrupt_path(_id),
            tmp.join('foo', '.dmedia', 'corrupt', _id)
        )

        # Test to make sure hashes are getting checked with check_id():
        bad = 'NWBNVXVK5..GIOW7MYR4K3KA5K22W7NW'
        with self.assertRaises(filestore.IDError) as cm:
            fs.corrupt_path(bad)
        self.assertIs(cm.exception.id, bad)

    def test_random_tmp_path(self):
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)
        filename = fs.random_tmp_path()
        self.assertEqual(path.dirname(filename), fs.tmp)
        name = path.basename(filename)
        self.assertEqual(len(name), 24)
        self.assertEqual(name,
            b32encode(b32decode(name.encode('utf-8'))).decode('utf-8')
        )
        self.assertEqual(path.abspath(filename), filename)

    def test_exists(self):
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)

        id1 = random_id()

        # File doesn't exist
        self.assertFalse(fs.exists(id1))

        # File exists
        tmp.write(b'Novacut', '.dmedia', 'files', id1[:2], id1[2:])
        self.assertTrue(fs.exists(id1))

        # Not file:
        id2 = random_id()
        tmp.makedirs('.dmedia', 'files', id2[:2], id2[2:])
        self.assertTrue(path.isdir(fs.path(id2)))
        self.assertFalse(fs.exists(id2))

        # Empty file
        id3 = random_id()
        f = fs.path(id3)
        assert not path.exists(f)
        open(f, 'wb').close()
        assert path.isfile(f)
        self.assertFalse(fs.exists(id3))

    def test_stat(self):
        # File doesn't exist
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)
        _id = random_id()
        with self.assertRaises(filestore.FileNotFound) as cm:
            st = fs.stat(_id)
        self.assertEqual(cm.exception.id, _id)
        self.assertIs(cm.exception.store, fs)
        self.assertEqual(
            str(cm.exception),
            '{!r} not in {!r}'.format(_id, fs)
        )

        # File is a symlink:
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)
        file = random_id()
        link = random_id()
        open(fs.path(file), 'wb').write(b'Novacut')
        os.symlink(fs.path(file), fs.path(link))
        assert path.isfile(fs.path(link))
        assert path.islink(fs.path(link))
        with self.assertRaises(filestore.FileNotFound) as cm:
            st = fs.stat(link)
        self.assertEqual(cm.exception.id, link)
        self.assertIs(cm.exception.store, fs)
        self.assertEqual(
            str(cm.exception),
            '{!r} not in {!r}'.format(link, fs)
        )

        # File is a directory
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)
        _id = random_id()
        os.mkdir(fs.path(_id))
        assert path.isdir(fs.path(_id))
        with self.assertRaises(filestore.FileNotFound) as cm:
            st = fs.stat(_id)
        self.assertEqual(cm.exception.id, _id)
        self.assertIs(cm.exception.store, fs)
        self.assertEqual(
            str(cm.exception),
            '{!r} not in {!r}'.format(_id, fs)
        )

        # Empty file
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)
        _id = random_id()
        open(fs.path(_id), 'wb').close()
        assert path.isfile(fs.path(_id))
        assert not path.islink(fs.path(_id))
        assert path.getsize(fs.path(_id)) == 0
        with self.assertRaises(filestore.FileNotFound) as cm:
            st = fs.stat(_id)
        self.assertEqual(cm.exception.id, _id)
        self.assertIs(cm.exception.store, fs)
        self.assertEqual(
            str(cm.exception),
            '{!r} not in {!r}'.format(_id, fs)
        )

        # Valid file
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)
        _id = random_id()
        open(fs.path(_id), 'wb').write(b'Novacut')
        st = fs.stat(_id)
        self.assertIsInstance(st, filestore.Stat)
        self.assertEqual(st.id, _id)
        self.assertEqual(st.size, 7)
        self.assertEqual(st.mtime, path.getmtime(fs.path(_id)))
        self.assertEqual(st.name, fs.path(_id))

    def test_open(self):
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)

        _id = random_id()

        # File doesn't exist
        with self.assertRaises(filestore.FileNotFound) as cm:
            fp = fs.open(_id)
        self.assertEqual(cm.exception.id, _id)
        self.assertIs(cm.exception.store, fs)

        # File exists
        data = b'ABC' * 1000
        canonical = fs.path(_id)
        open(canonical, 'wb').write(data)
        fp = fs.open(_id)
        self.assertIsInstance(fp, io.BufferedReader)
        self.assertEqual(fp.mode, 'rb')
        self.assertEqual(fp.tell(), 0)
        self.assertEqual(fp.name, canonical)
        self.assertEqual(fp.read(), data)

    def test_verify(self):
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)

        canonical = fs.path(ID)
        corrupt = fs.corrupt_path(ID)

        # File doesn't exist
        with self.assertRaises(filestore.FileNotFound) as cm:
            ch = fs.verify(ID)
        self.assertEqual(cm.exception.id, ID)
        self.assertIs(cm.exception.store, fs)

        # File exists:
        fp = open(canonical, 'wb')
        for leaf in LEAVES:
            fp.write(leaf)
        fp.close()

        self.assertEqual(fs.verify(ID), CH)  # return_fp=False
        self.assertEqual(fs.verify(ID, False), CH)  # return_fp=False

        fp = fs.verify(ID, True)  # return_fp=True
        self.assertEqual(fp.mode, 'rb')
        self.assertEqual(fp.tell(), 0)
        self.assertEqual(fp.name, canonical)

        # File exists and is corrupted
        fp = open(canonical, 'wb')
        for leaf in LEAVES:
            fp.write(leaf)
        fp.write(b'F')
        fp.close()
        ch = filestore.hash_fp(open(canonical, 'rb'))
        self.assertNotEqual(ch.id, ID)

        self.assertTrue(path.isfile(canonical))
        self.assertFalse(path.exists(corrupt))

        with self.assertRaises(filestore.CorruptFile) as cm:
            ch = fs.verify(ID)
        self.assertEqual(cm.exception.id, ID)
        self.assertIs(cm.exception.store, fs)
        self.assertEqual(cm.exception.kw, {'bad_id': ch.id})

        self.assertFalse(path.exists(canonical))
        self.assertTrue(path.isfile(corrupt))

    def test_verify2(self):
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)
        obj = misc.load_test_vectors()
        for proto in PROTOCOLS:
            leaves = misc.build_test_leaves(proto.leaf_size)
            vectors = obj[str(proto.digest_b32len)]
            for (key, value) in vectors['root_hashes'].items():
                name = fs.path(value)
                assert not path.exists(name)
                fp = open(name, 'wb')
                for L in key:
                    fp.write(leaves[L])
                fp.close()
                ch = fs.verify(value)
                self.assertIsInstance(ch, filestore.ContentHash)
                self.assertEqual(ch.id, value)
                self.assertEqual(ch.file_size, path.getsize(name))
                self.assertEqual(ch.leaf_hashes,
                    b''.join(
                        b32loads(vectors['leaf_hashes'][L][i])
                        for (i, L) in enumerate(key)
                    )
                )

    def test_verify_iter(self):
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)
        canonical = fs.path(ID)
        corrupt = fs.corrupt_path(ID)

        # File doesn't exist
        with self.assertRaises(filestore.FileNotFound) as cm:
            for leaf in fs.verify_iter(ID, FILE_SIZE, LEAF_HASHES):
                pass
        self.assertEqual(cm.exception.id, ID)
        self.assertIs(cm.exception.store, fs)

        # File exists
        fp = open(canonical, 'wb')
        for leaf in LEAVES:
            fp.write(leaf)
        fp.close()

        j = 0
        for (i, leaf) in fs.verify_iter(ID, FILE_SIZE, LEAF_HASHES):
            self.assertEqual(leaf, LEAVES[i])
            self.assertEqual(i, j)
            j += 1
        self.assertEqual(i, 2)

        # File exists and file size is 1 byte too small:
        fp = open(canonical, 'wb')
        fp.write(LEAVES[0])
        fp.write(LEAVES[1])
        fp.write(LEAVES[2][:-1])
        fp.close()
        self.assertTrue(path.isfile(canonical))
        self.assertFalse(path.exists(corrupt))

        with self.assertRaises(filestore.CorruptFile) as cm:
            for leaf in fs.verify_iter(ID, FILE_SIZE, LEAF_HASHES):
                pass
        self.assertEqual(cm.exception.id, ID)
        self.assertIs(cm.exception.store, fs)
        self.assertEqual(
            cm.exception.kw,
            {'file_size': FILE_SIZE, 'bad_file_size': FILE_SIZE - 1}
        )
        self.assertFalse(path.exists(canonical))
        self.assertTrue(path.isfile(corrupt))
        os.remove(corrupt)

        # File exists and file size is 1 byte too large:
        fp = open(canonical, 'wb')
        fp.write(LEAVES[0])
        fp.write(LEAVES[1])
        fp.write(LEAVES[2] + b'c')
        fp.close()
        self.assertTrue(path.isfile(canonical))
        self.assertFalse(path.exists(corrupt))

        with self.assertRaises(filestore.CorruptFile) as cm:
            for leaf in fs.verify_iter(ID, FILE_SIZE, LEAF_HASHES):
                pass
        self.assertEqual(cm.exception.id, ID)
        self.assertIs(cm.exception.store, fs)
        self.assertEqual(
            cm.exception.kw,
            {'file_size': FILE_SIZE, 'bad_file_size': FILE_SIZE + 1}
        )
        self.assertFalse(path.exists(canonical))
        self.assertTrue(path.isfile(corrupt))
        os.remove(corrupt)

        # File exists has corruption on 2nd leaf:
        fp = open(canonical, 'wb')
        fp.write(LEAVES[0])
        fp.write(LEAVES[1][:-1] + b'f')
        fp.write(LEAVES[2])
        fp.close()
        self.assertTrue(path.isfile(canonical))
        self.assertFalse(path.exists(corrupt))

        with self.assertRaises(filestore.CorruptFile) as cm:
            for leaf in fs.verify_iter(ID, FILE_SIZE, LEAF_HASHES):
                pass
        self.assertEqual(cm.exception.id, ID)
        self.assertIs(cm.exception.store, fs)
        self.assertEqual(
            cm.exception.kw,
            {
                'leaf_index': 1,
                'leaf_hash': filestore.hash_leaf(1, LEAVES[1]),
                'bad_leaf_hash': filestore.hash_leaf(1, LEAVES[1][:-1] + b'f'),
            }
        )
        self.assertFalse(path.exists(canonical))
        self.assertTrue(path.isfile(corrupt))
        os.remove(corrupt)

    def test_verify_iter2(self):
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)

        canonical = fs.path(ID)
        corrupt = fs.corrupt_path(ID)

        # File doesn't exist
        with self.assertRaises(filestore.FileNotFound) as cm:
            for leaf in fs.verify_iter2(ID):
                pass
        self.assertEqual(cm.exception.id, ID)
        self.assertIs(cm.exception.store, fs)

        # File exists:
        fp = open(canonical, 'wb')
        for leaf in LEAVES:
            fp.write(leaf)
        fp.close()

        self.assertEqual(
            tuple(fs.verify_iter2(ID)),
            tuple(enumerate(LEAVES))
        )

        # File exists and is corrupted
        fp = open(canonical, 'wb')
        for leaf in LEAVES:
            fp.write(leaf)
        fp.write(b'F')
        fp.close()
        ch = filestore.hash_fp(open(canonical, 'rb'))
        self.assertNotEqual(ch.id, ID)

        self.assertTrue(path.isfile(canonical))
        self.assertFalse(path.exists(corrupt))

        with self.assertRaises(filestore.CorruptFile) as cm:
            for leaf in fs.verify_iter2(ID):
                pass
        self.assertEqual(cm.exception.id, ID)
        self.assertEqual(cm.exception.store, fs)
        self.assertEqual(cm.exception.kw, {'bad_id': ch.id})

        self.assertFalse(path.exists(canonical))
        self.assertTrue(path.isfile(corrupt))

    def test_content_md5(self):
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)

        canonical = fs.path(ID)
        corrupt = fs.corrupt_path(ID)

        # File doesn't exist
        with self.assertRaises(filestore.FileNotFound) as cm:
            md5 = fs.content_md5(ID)
        self.assertEqual(cm.exception.id, ID)
        self.assertIs(cm.exception.store, fs)

        # File exists:
        fp = open(canonical, 'wb')
        for leaf in LEAVES:
            fp.write(leaf)
        fp.close()

        self.assertEqual(
            fs.content_md5(ID),
            ('99ca2a74521ad7825768bbfe7fe0dc49', 'mcoqdFIa14JXaLv+f+DcSQ==')
        )

        # File exists and is corrupted
        fp = open(canonical, 'wb')
        for leaf in LEAVES:
            fp.write(leaf)
        fp.write(b'F')
        fp.close()
        ch = filestore.hash_fp(open(canonical, 'rb'))
        self.assertNotEqual(ch.id, ID)

        self.assertTrue(path.isfile(canonical))
        self.assertFalse(path.exists(corrupt))

        with self.assertRaises(filestore.CorruptFile) as cm:
            md5 = fs.content_md5(ID)
        self.assertEqual(cm.exception.id, ID)
        self.assertEqual(cm.exception.store, fs)
        self.assertEqual(cm.exception.kw, {'bad_id': ch.id})

        self.assertFalse(path.exists(canonical))
        self.assertTrue(path.isfile(corrupt))

    def test_remove(self):
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)
        _id = random_id()
        canonical = fs.path(_id)

        # File doesn't exist
        with self.assertRaises(OSError) as cm:
            fs.remove(_id)
        self.assertEqual(cm.exception.errno, 2)

        # File exists
        open(canonical, 'wb').write(b'hello')
        self.assertTrue(path.isfile(canonical))
        fs.remove(_id)
        self.assertFalse(path.exists(canonical))

    def test_purge_tmp(self):
        tmp = TempDir()
        t = tmp.join('.dmedia', 'tmp')
        fs = filestore.FileStore(tmp.dir)
        self.assertTrue(path.isdir(t))

        # Test when nothing is in tmp:
        self.assertEqual(os.listdir(t), [])
        self.assertEqual(fs.purge_tmp(), [])
        self.assertEqual(os.listdir(t), [])

        # Test when only regular files are in tmp:
        purged = []
        for i in range(20):
            size = random.randint(1, 1024 * 1024)
            fp = fs.allocate_tmp(size)
            purged.append(
                filestore.File(fp.name, size, path.getmtime(fp.name))
            )
        purged.sort(key=lambda f: f.name)
        self.assertEqual(fs.purge_tmp(), purged)
        self.assertEqual(os.listdir(t), [])

        # Test that directories aren't traversed:
        nope = []
        for d in ['aye', 'bee', 'see']:
            for n in 'abcd':
                nope.append(tmp.touch('.dmedia', 'tmp', d, n))
        self.assertEqual(sorted(os.listdir(t)), ['aye', 'bee', 'see'])
        self.assertEqual(fs.purge_tmp(), [])
        self.assertEqual(sorted(os.listdir(t)), ['aye', 'bee', 'see'])
        for f in nope:
            self.assertTrue(path.isfile(f))

        # Test that symlinks are ignored:
        tmp2 = TempDir()
        fs2 = filestore.FileStore(tmp2.dir)
        names = ['aye', 'bee', 'see']
        for i in range(10):
            size = random.randint(1, 1024 * 1024)
            fp = fs2.allocate_tmp(size)
            name = path.basename(fp.name)
            link = path.join(t, name)
            os.symlink(fp.name, link)
            names.append(name)
            nope.append(link)
        names.sort()
        self.assertEqual(sorted(os.listdir(t)), names)
        for f in nope:
            self.assertTrue(path.isfile(f))

        # Add some regular files, including zero-sized files:
        purged = []
        for i in range(20):
            size = random.randint(1, 1024 * 1024)
            fp = fs.allocate_tmp(size)
            purged.append(
                filestore.File(fp.name, size, path.getmtime(fp.name))
            )
        for i in range(10):
            fp = fs.allocate_tmp()
            purged.append(
                filestore.File(fp.name, 0, path.getmtime(fp.name))
            )
        purged.sort(key=lambda f: f.name)
        self.assertEqual(fs.purge_tmp(), purged)
        self.assertEqual(sorted(os.listdir(t)), names)
        for f in nope:
            self.assertTrue(path.isfile(f))

    def test_allocate_tmp(self):
        tmp = TempDir()
        t = tmp.join('.dmedia', 'tmp')
        fs = filestore.FileStore(tmp.dir)
        self.assertTrue(path.isdir(t))

        # Test with size=None
        fp1 = fs.allocate_tmp()
        self.assertIsInstance(fp1, io.BufferedWriter)
        self.assertEqual(fp1.mode, 'wb')
        self.assertEqual(os.fstat(fp1.fileno()).st_size, 0)
        self.assertEqual(stat.S_IMODE(os.fstat(fp1.fileno()).st_mode), 0o666)
        self.assertTrue(fp1.name.startswith(t + os.sep))

        # Test with size=2311
        fp2 = fs.allocate_tmp(2311)
        self.assertIsInstance(fp2, io.BufferedWriter)
        self.assertEqual(fp2.mode, 'wb')
        self.assertEqual(os.fstat(fp2.fileno()).st_size, 2311)
        self.assertEqual(stat.S_IMODE(os.fstat(fp2.fileno()).st_mode), 0o666)
        self.assertTrue(fp2.name.startswith(t + os.sep))
        self.assertNotEqual(fp2.name, fp1.name)

        # Test with bad size type:
        with self.assertRaises(TypeError) as cm:
            fp = fs.allocate_tmp(2311.0)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('size', int, float, 2311.0)
        )

        # Test with bad size value:
        with self.assertRaises(ValueError) as cm:
            fp = fs.allocate_tmp(0)
        self.assertEqual(
            str(cm.exception),
            'Need size >= 1; got size=0'
        )

        # But make sure that size=1 works:
        fp3 = fs.allocate_tmp(1)
        self.assertIsInstance(fp3, io.BufferedWriter)
        self.assertEqual(fp3.mode, 'wb')
        self.assertEqual(os.fstat(fp3.fileno()).st_size, 1)
        self.assertEqual(stat.S_IMODE(os.fstat(fp3.fileno()).st_mode), 0o666)
        self.assertTrue(fp3.name.startswith(t + os.sep))
        self.assertNotEqual(fp3.name, fp1.name)
        self.assertNotEqual(fp3.name, fp2.name)

    def test_allocate_partial(self):
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)
        _id = random_id()
        filename = tmp.join('.dmedia', 'partial', _id)

        # Test when file dosen't yet exist
        fp = fs.allocate_partial(2311, _id)
        self.assertEqual(fp.name, filename)
        self.assertIsInstance(fp, io.BufferedWriter)
        self.assertEqual(fp.mode, 'wb')
        self.assertEqual(os.fstat(fp.fileno()).st_size, 2311)
        self.assertEqual(stat.S_IMODE(os.fstat(fp.fileno()).st_mode), 0o666)

        fp.write(b'a' * 3141)  # Extend past size
        fp.close()

        # Test with pre-existing file larger than size:
        self.assertEqual(os.stat(filename).st_size, 3141)
        fp = fs.allocate_partial(2311, _id)
        self.assertEqual(fp.name, filename)
        self.assertIsInstance(fp, io.BufferedRandom)
        self.assertIn(fp.mode, ('r+b', 'rb+'))
        self.assertEqual(os.fstat(fp.fileno()).st_size, 2311)
        self.assertEqual(stat.S_IMODE(os.fstat(fp.fileno()).st_mode), 0o666)
        fp.close()

        open(filename, 'wb').write(b'a' * 200)  # Truncate to smaller than size

        # Test with pre-existing file smaller than size:
        self.assertEqual(os.stat(filename).st_size, 200)
        fp = fs.allocate_partial(2311, _id)
        self.assertEqual(fp.name, filename)
        self.assertIsInstance(fp, io.BufferedRandom)
        self.assertIn(fp.mode, ('r+b', 'rb+'))
        self.assertEqual(os.fstat(fp.fileno()).st_size, 2311)
        self.assertEqual(stat.S_IMODE(os.fstat(fp.fileno()).st_mode), 0o666)


        _id = random_id()  # We'll use a new ID for below
        filename = tmp.join('.dmedia', 'partial', _id)

        # Test with bad size type:
        with self.assertRaises(TypeError) as cm:
            fp = fs.allocate_partial(2311.0, _id)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('size', int, float, 2311.0)
        )

        # Test with bad size value:
        with self.assertRaises(ValueError) as cm:
            fp = fs.allocate_partial(0, _id)
        self.assertEqual(
            str(cm.exception),
            'Need size >= 1; got size=0'
        )

        # But make sure that size=1 works:
        fp = fs.allocate_partial(1, _id)
        self.assertEqual(fp.name, filename)
        self.assertIsInstance(fp, io.BufferedWriter)
        self.assertEqual(fp.mode, 'wb')
        self.assertEqual(os.fstat(fp.fileno()).st_size, 1)
        self.assertEqual(stat.S_IMODE(os.fstat(fp.fileno()).st_mode), 0o666)
        fp.close()
        fp = fs.allocate_partial(1, _id)
        self.assertEqual(fp.name, filename)
        self.assertIsInstance(fp, io.BufferedRandom)
        self.assertIn(fp.mode, ('r+b', 'rb+'))
        self.assertEqual(os.fstat(fp.fileno()).st_size, 1)
        self.assertEqual(stat.S_IMODE(os.fstat(fp.fileno()).st_mode), 0o666)

    def test_move_to_canonical(self):
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)
        _id = random_id()
        dst = fs.path(_id)

        # Test with wrong src_fp type
        with self.assertRaises(TypeError) as cm:
            fs.move_to_canonical(17, _id)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('src_fp', io.BufferedIOBase, int, 17)
        )
        self.assertFalse(path.exists(dst))

        # Test with src_fp opened in wrong mode
        bad = tmp.touch('.dmedia', 'bad1.mov')
        for mode in ('r', 'w', 'r+'):
            fp = open(bad, mode)
            with self.assertRaises(TypeError) as cm:
                fs.move_to_canonical(fp, _id)
            self.assertEqual(
                str(cm.exception),
                TYPE_ERROR.format('src_fp', io.BufferedIOBase, type(fp), fp)
            )
            self.assertFalse(path.exists(dst))
            self.assertFalse(fp.closed)

        # Test when filename isn't contained in basedir (to ensure that
        # FileStore.catch_traversal() is used to validate src_fp.name)
        bad = tmp.touch('.dmedia', '..', 'bad2.mov')
        fp = open(bad, 'rb')
        with self.assertRaises(filestore.PathTraversal) as cm:
            fs.move_to_canonical(fp, _id)
        self.assertEqual(cm.exception.abspath, tmp.join('bad2.mov'))
        self.assertIs(cm.exception.store, fs)
        self.assertFalse(path.exists(dst))
        self.assertFalse(fp.closed)

        # Test that _id is validated with check_id()
        good = tmp.write(b'yup', '.dmedia', 'tmp', 'good.mov')
        os.chmod(good, 0o660)
        fp = open(good, 'rb')
        b32 = 'NWBNVXVK5DQGIOW7MYR4K3KA'
        with self.assertRaises(filestore.IDError) as cm:
            fs.move_to_canonical(fp, b32)
        self.assertIs(cm.exception.id, b32)
        self.assertFalse(path.exists(dst))
        self.assertFalse(fp.closed)

        # Test when src_fp is closed
        fp.close()
        with self.assertRaises(ValueError) as cm:
            fs.move_to_canonical(fp, _id)
        self.assertEqual(
            str(cm.exception),
            'src_fp is closed, must be open: {!r}'.format(fp)
        )
        self.assertFalse(path.exists(dst))
        self.assertTrue(fp.closed)

        # Test when it's all good
        fp = open(good, 'rb')
        self.assertEqual(stat.S_IMODE(os.stat(good).st_mode), 0o660)
        self.assertEqual(fs.move_to_canonical(fp, _id), dst)
        self.assertTrue(fp.closed)
        self.assertFalse(path.exists(good))
        self.assertTrue(path.isfile(dst))
        self.assertEqual(stat.S_IMODE(os.stat(dst).st_mode), 0o444)
        self.assertEqual(open(dst, 'rb').read(), b'yup')

    def test_move_to_corrupt(self):
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)
        _id = random_id()
        corrupt = fs.corrupt_path(_id)
        canonical = fs.path(_id)

        # Test with wrong fp type
        with self.assertRaises(TypeError) as cm:
            fs.move_to_corrupt(17, _id)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('src_fp', io.BufferedReader, int, 17)
        )
        self.assertFalse(path.exists(corrupt))

        # Test with fp opened 'r' mode:
        bad = tmp.touch('.dmedia', 'bad1.mov')
        fp = open(bad, 'r')
        with self.assertRaises(TypeError) as cm:
            fs.move_to_corrupt(fp, _id)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('src_fp', io.BufferedReader, io.TextIOWrapper, fp)
        )
        self.assertFalse(path.exists(corrupt))
        self.assertFalse(fp.closed)

        # Test with fp opened 'rb+' mode:
        fp = open(bad, 'rb+')
        with self.assertRaises(TypeError) as cm:
            fs.move_to_corrupt(fp, _id)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('src_fp', io.BufferedReader, io.BufferedRandom, fp)
        )
        self.assertFalse(path.exists(corrupt))
        self.assertFalse(fp.closed)

        # Test when filename isn't contained in basedir (to ensure that
        # FileStore.catch_traversal() is used to validate tmp_fp.name)
        bad = tmp.touch('.dmedia', '..', 'bad2.mov')
        fp = open(bad, 'rb')
        with self.assertRaises(filestore.PathTraversal) as cm:
            fs.move_to_corrupt(fp, _id)
        self.assertEqual(cm.exception.abspath, tmp.join('bad2.mov'))
        self.assertIs(cm.exception.store, fs)
        self.assertFalse(path.exists(corrupt))
        self.assertFalse(fp.closed)

        # Test that _id is validated with check_id()
        good = tmp.write(b'yup', '.dmedia', 'tmp', 'good.mov')
        fp = open(good, 'rb')
        b32 = 'NWBNVXVK5DQGIOW7MYR4K3KA'
        with self.assertRaises(filestore.IDError) as cm:
            fs.move_to_corrupt(fp, b32)
        self.assertIs(cm.exception.id, b32)
        self.assertFalse(path.exists(corrupt))
        self.assertFalse(fp.closed)

        # Test when fp is closed
        fp.close()
        with self.assertRaises(ValueError) as cm:
            fs.move_to_corrupt(fp, _id)
        self.assertEqual(
            str(cm.exception),
            'src_fp is closed, must be open: {!r}'.format(fp)
        )
        self.assertFalse(path.exists(corrupt))
        self.assertTrue(fp.closed)

        # Test when it's all good
        fp = open(good, 'rb')
        e = fs.move_to_corrupt(fp, _id)
        self.assertIsInstance(e, filestore.CorruptFile)
        self.assertEqual(e.id, _id)
        self.assertIs(e.store, fs)
        self.assertEqual(e.kw, {})
        self.assertTrue(fp.closed)
        self.assertFalse(path.exists(good))
        self.assertTrue(path.isfile(corrupt))
        self.assertEqual(open(corrupt, 'rb').read(), b'yup')

    def test_verify_and_move(self):
        # Test when it's all good
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)
        src = fs.partial_path(ID)
        write_sample_file(src)
        src_fp = open(src, 'rb')
        self.assertFalse(fs.exists(ID))
        self.assertEqual(fs.verify_and_move(src_fp, ID), CH)
        self.assertTrue(fs.exists(ID))
        self.assertEqual(fs.verify(ID), CH)

        # Test when tmp_fp.name != partial_path(_id)
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)
        src = path.join(fs.tmp, 'foo.mov')
        write_sample_file(src)
        src_fp = open(src, 'rb')
        with self.assertRaises(ValueError) as cm:
            fs.verify_and_move(src_fp, ID)
        self.assertEqual(
            str(cm.exception),
            'bad partial_path() for {!r}: {!r}'.format(ID, src)
        )
        self.assertFalse(fs.exists(ID))

        # Test when partial has wrong content
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)
        src = fs.partial_path(ID)
        src_fp = open(src, 'wb')
        for leaf in LEAVES:
            src_fp.write(leaf)
        src_fp.write(b'F')
        src_fp.close()
        ch = filestore.hash_fp(open(src, 'rb'))
        self.assertNotEqual(ch.id, ID)
        src_fp = open(src, 'rb')
        with self.assertRaises(ValueError) as cm:
            fs.verify_and_move(src_fp, ID)
        self.assertEqual(
            str(cm.exception),
            'expected {!r}, computed {!r}'.format(ID, ch.id)
        )
        self.assertFalse(fs.exists(ID))

    def test_hash_and_move(self):
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)
        src = path.join(fs.tmp, 'foo.mov')
        write_sample_file(src)
        src_fp = open(src, 'rb')
        self.assertFalse(fs.exists(ID))
        self.assertEqual(fs.hash_and_move(src_fp), CH)
        self.assertTrue(fs.exists(ID))
        self.assertEqual(fs.verify(ID), CH)

        # Test David's use case:
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)
        tmp_fp = fs.allocate_tmp()
        write_sample_file(tmp_fp)
        self.assertFalse(fs.exists(ID))
        self.assertEqual(fs.hash_and_move(tmp_fp), CH)
        self.assertTrue(fs.exists(ID))
        self.assertEqual(fs.verify(ID), CH)

    def test_import_file(self):
        tmp = TempDir()
        src = tmp.join('movie.mov')
        write_sample_file(src)

        fs = filestore.FileStore(tmp.dir)
        dst = fs.path(ID)
        self.assertTrue(path.isfile(src))
        self.assertFalse(path.exists(dst))

        src_fp = open(src, 'rb')
        self.assertEqual(fs.import_file(src_fp), CH)
        self.assertTrue(path.isfile(src))
        self.assertTrue(path.isfile(dst))

        self.assertEqual(filestore.hash_fp(open(dst, 'rb')), CH)

    def test_copy(self):
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)

        tmp1 = TempDir()
        tmp2 = TempDir()
        fs1 = filestore.FileStore(tmp1.dir)
        fs2 = filestore.FileStore(tmp2.dir)

        with self.assertRaises(TypeError) as cm:
            fs.copy(ID)
        self.assertEqual(
            str(cm.exception),
            'FileStore.copy() takes at least one destination filestore'
        )

        # Bad type in filestores
        with self.assertRaises(TypeError) as cm:
            fs.copy(ID, 17, fs2)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('filestores[0]', filestore.FileStore, int, 17)
        )
        with self.assertRaises(TypeError) as cm:
            fs.copy(ID, fs1, 18.0)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('filestores[1]', filestore.FileStore, float, 18.0)
        )

        # Same parentdir in filestores
        fs_same = filestore.FileStore(tmp.dir)
        with self.assertRaises(ValueError) as cm:
            fs.copy(ID, fs_same, fs2)
        self.assertEqual(
            str(cm.exception),
            'wont copy into filestores[0], same parentdir as {!r}'.format(fs)
        )
        with self.assertRaises(ValueError) as cm:
            fs.copy(ID, fs1, fs_same)
        self.assertEqual(
            str(cm.exception),
            'wont copy into filestores[1], same parentdir as {!r}'.format(fs)
        )

        # src file doesn't exist:
        with self.assertRaises(filestore.FileNotFound) as cm:
            fs.copy(ID, fs1, fs2)
        self.assertEqual(cm.exception.id, ID)
        self.assertIs(cm.exception.store, fs)

        # File exists and is corrupted
        canonical = fs.path(ID)
        fp = open(canonical, 'wb')
        for leaf in LEAVES:
            fp.write(leaf)
        fp.write(b'F')
        fp.close()
        ch = filestore.hash_fp(open(canonical, 'rb'))
        self.assertNotEqual(ch.id, ID)

        with self.assertRaises(filestore.CorruptFile) as cm:
            fs.copy(ID, fs1, fs2)
        self.assertEqual(cm.exception.id, ID)
        self.assertIs(cm.exception.store, fs)
        self.assertEqual(cm.exception.kw, {'bad_id': ch.id})
        self.assertFalse(path.exists(canonical))
        self.assertTrue(path.isfile(fs.corrupt_path(ID)))
        os.remove(fs.corrupt_path(ID))

        self.assertFalse(fs1.exists(ID))
        self.assertEqual(os.listdir(fs1.tmp), [])
        self.assertEqual(list(fs1), [])

        self.assertFalse(fs2.exists(ID))
        self.assertEqual(os.listdir(fs2.tmp), [])
        self.assertEqual(list(fs2), [])

        # File exists and is valid
        fp = open(canonical, 'wb')
        for leaf in LEAVES:
            fp.write(leaf)
        fp.close()
        self.assertEqual(fs.copy(ID, fs1, fs2), CH)

        self.assertTrue(fs1.exists(ID))
        self.assertEqual([s.id for s in fs1], [ID])
        self.assertEqual(os.listdir(fs1.tmp), [])

        self.assertTrue(fs2.exists(ID))
        self.assertEqual([s.id for s in fs2], [ID])
        self.assertEqual(os.listdir(fs2.tmp), [])

        fs1.verify(ID)
        fs2.verify(ID)
        fs.verify(ID)

        # File already exists in destination filestores
        self.assertEqual(fs.copy(ID, fs1, fs2), CH)

        self.assertTrue(fs1.exists(ID))
        self.assertEqual([s.id for s in fs1], [ID])
        self.assertEqual(os.listdir(fs1.tmp), [])

        self.assertTrue(fs2.exists(ID))
        self.assertEqual([s.id for s in fs2], [ID])
        self.assertEqual(os.listdir(fs2.tmp), [])

        fs1.verify(ID)
        fs2.verify(ID)
        fs.verify(ID)

