# filestore: dmedia hashing protocol and file layout
# Copyright (C) 2012 Novacut Inc
#
# This file is part of `filestore`.
#
# `filestore` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `filestore` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License along
# with `filestore`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Unit tests for `filestore.misc` module.
"""

from unittest import TestCase
import os
from os import path
import json
from hashlib import md5

import filestore
from filestore.protocols import MIN_LEAF_SIZE, Protocol, VERSION0, VERSION1
from filestore.protocols import b32dumps, b32loads
from filestore import misc


class TestFunctions(TestCase):
    def test_load_test_vectors(self):
        pkg = path.dirname(path.abspath(filestore.__file__))
        self.assertTrue(path.isdir(pkg))
        self.assertEqual(
            misc.TEST_VECTORS,
            path.join(pkg, 'data', 'test-vectors.json')
        )
        self.assertTrue(path.isfile(misc.TEST_VECTORS))
        self.assertEqual(
            misc.load_test_vectors(),
            json.loads(open(misc.TEST_VECTORS, 'r').read())
        )

    def test_build_test_leaves(self):
        obj = misc.build_test_leaves(MIN_LEAF_SIZE)
        self.assertEqual(obj,
            {
                'A': b'A',
                'B': b'B' * (MIN_LEAF_SIZE - 1),
                'C': b'C' * MIN_LEAF_SIZE,
            }
        )
        self.assertEqual(len(obj['A']), 1)
        self.assertEqual(len(obj['B']), MIN_LEAF_SIZE - 1)
        self.assertEqual(len(obj['C']), MIN_LEAF_SIZE)

        obj = misc.build_test_leaves(2 * MIN_LEAF_SIZE)
        self.assertEqual(obj,
            {
                'A': b'A',
                'B': b'B' * (2 * MIN_LEAF_SIZE - 1),
                'C': b'C' * (2 * MIN_LEAF_SIZE),
            }
        )
        self.assertEqual(len(obj['A']), 1)
        self.assertEqual(len(obj['B']), 2 * MIN_LEAF_SIZE - 1)
        self.assertEqual(len(obj['C']), 2 * MIN_LEAF_SIZE)

    def test_build_test_vectors(self):
        proto = Protocol(MIN_LEAF_SIZE, 200)
        leaves = misc.build_test_leaves(proto.leaf_size)
        A0 = proto.hash_leaf(0, leaves['A'])
        A1 = proto.hash_leaf(1, leaves['A'])
        B0 = proto.hash_leaf(0, leaves['B'])
        B1 = proto.hash_leaf(1, leaves['B'])
        C0 = proto.hash_leaf(0, leaves['C'])
        C1 = proto.hash_leaf(1, leaves['C'])
        self.assertEqual(
            misc.build_test_vectors(proto),
            {
                'leaf_hashes': {
                    'A': [b32dumps(A0), b32dumps(A1)],
                    'B': [b32dumps(B0), b32dumps(B1)],
                    'C': [b32dumps(C0), b32dumps(C1)],
                },
                'root_hashes': {
                    'A': proto.hash_root(1, A0),
                    'B': proto.hash_root(MIN_LEAF_SIZE - 1, B0),
                    'C': proto.hash_root(MIN_LEAF_SIZE, C0),
                    'CA': proto.hash_root(MIN_LEAF_SIZE + 1, C0 + A1),
                    'CB': proto.hash_root(2 * MIN_LEAF_SIZE - 1, C0 + B1),
                    'CC': proto.hash_root(2 * MIN_LEAF_SIZE, C0 + C1),
                }
            }
        )

        self.assertEqual(
            misc.load_test_vectors(),
            {
                '48': misc.build_test_vectors(VERSION0),
                '56': misc.build_test_vectors(VERSION1),
            }
        )

    def test_get_test_integers(self):
        self.assertEqual(
            misc.get_test_integers(100),
            [0, 1, 99, 100, 101, 199, 200]
        )
        self.assertEqual(
            misc.get_test_integers(500),
            [0, 1, 499, 500, 501, 999, 1000]
        )

    def test_build_test_md5(self):
        A = b'A'
        B = b'B' * 99
        C = b'C' * 100
        self.assertEqual(
            misc.build_test_md5(100),
            {
                'integers': {
                    '0': md5(b'0').hexdigest(),
                    '1': md5(b'1').hexdigest(),
                    '99': md5(b'99').hexdigest(),
                    '100': md5(b'100').hexdigest(),
                    '101': md5(b'101').hexdigest(),
                    '199': md5(b'199').hexdigest(),
                    '200': md5(b'200').hexdigest(),
                },
                'files': {
                    'A': md5(A).hexdigest(),
                    'B': md5(B).hexdigest(),
                    'C': md5(C).hexdigest(),
                    'CA': md5(C + A).hexdigest(),
                    'CB': md5(C + B).hexdigest(),
                    'CC': md5(C + C).hexdigest(),
                }
            }   
        )


class TestTempFileStore(TestCase):
    def test_init(self):
        fs = misc.TempFileStore()
        self.assertIsInstance(fs, filestore.FileStore)
        self.assertTrue(fs.parentdir.startswith('/tmp/TempFileStore.'))
        self.assertIsNone(fs.id)
        self.assertEqual(fs.copies, 0)

        fs = misc.TempFileStore('foobar', 2)
        self.assertIsInstance(fs, filestore.FileStore)
        self.assertTrue(fs.parentdir.startswith('/tmp/TempFileStore.'))
        self.assertEqual(fs.id, 'foobar')
        self.assertEqual(fs.copies, 2)

        fs = misc.TempFileStore(_id='hooray', copies=3)
        self.assertIsInstance(fs, filestore.FileStore)
        self.assertTrue(fs.parentdir.startswith('/tmp/TempFileStore.'))
        self.assertEqual(fs.id, 'hooray')
        self.assertEqual(fs.copies, 3)

    def test_del(self):
        fs = misc.TempFileStore()
        parentdir = fs.parentdir
        self.assertTrue(path.isdir(parentdir))
        fs = None
        self.assertFalse(path.exists(parentdir))
