# filestore: dmedia hashing protocol and file layout
# Copyright (C) 2012 Novacut Inc
#
# This file is part of `filestore`.
#
# `filestore` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `filestore` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License along
# with `filestore`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Unit tests for `filestore.protocols` module.
"""

from unittest import TestCase
import os
from base64 import b32encode, b32decode
from hashlib import md5

from skein import skein512

from filestore.protocols import b32dumps, b32loads
from filestore import protocols


MiB = 1024 * 1024
TwoMiB = 2 * MiB
EightMiB = 8 * MiB
COUNT = 50 * 1000
TYPE_ERROR = '{}: need a {!r}; got a {!r}: {!r}'

A = b'A'
B = b'B' * (EightMiB - 1)
C = b'C' * EightMiB

A0_B32 = 'XZ5I6KJTUSOIWVCEBOKUELTADZUXNHOAYO77NKKHWCIW3HYGYOPMX5JN'
A1_B32 = 'TEC7754ZNM26MTM6YQFI6TMVTTK4RKQEMPAGT2ROQZUBPUIHSJU2DDR3'
B0_B32 = 'P67PVKU3SCCQHNIRMR2Z5NICEMIP36WCFJG4AW6YBAE6UI4K6BVLY3EI'
B1_B32 = 'ZIFO5S2OYYPZAUN6XQWTWZGCDATXCGR2JYN7UIAX54WMVWETMIUFG7WM'
C0_B32 = 'RW2GJFIGPQF5WLR53UAK77TPHNRFKMUBYRB23JFS4G2RFRRNHW6OX4CR'
C1_B32 = 'XBVLPYBUX6QD2DKPJTYVUXT23K3AAUAW5J4RMQ543NQNDAHORQJ7GBDE'


class TestConstants(TestCase):
    def test_max_file_size(self):
        # For now, constrain to JavaScript integer limit (2**53):
        self.assertEqual(protocols.MAX_FILE_SIZE, 9007199254740992)
        self.assertEqual(protocols.MAX_FILE_SIZE, 2**53)

    def test_min_leaf_size(self):
        self.assertEqual(protocols.MIN_LEAF_SIZE, 2 * MiB)
        self.assertEqual(protocols.MIN_LEAF_SIZE,
            protocols.MAX_FILE_SIZE // 2**32
        )


class TestFunctions(TestCase):
    def test_b32dumps(self):
        self.assertEqual(b32dumps(b'\x00\x00\x00\x00\x00'), 'AAAAAAAA')
        self.assertEqual(b32dumps(b'\xff\xff\xff\xff\xff'), '77777777')
        self.assertEqual(b32dumps(b'\x00' * 35), 'A' * 56)
        self.assertEqual(b32dumps(b'\xff' * 35), '7' * 56)
        for i in range(1000):
            data = os.urandom(15)
            self.assertEqual(
                b32dumps(data),
                b32encode(data).decode('utf-8')
            )

    def test_b32loads(self):
        self.assertEqual(b32loads('AAAAAAAA'), b'\x00\x00\x00\x00\x00')
        self.assertEqual(b32loads('77777777'), b'\xff\xff\xff\xff\xff')
        self.assertEqual(b32loads('A' * 56), b'\x00' * 35)
        self.assertEqual(b32loads('7' * 56), b'\xff' * 35)
        for i in range(1000):
            data = os.urandom(15)
            b32 = b32encode(data)
            self.assertEqual(b32loads(b32.decode('utf-8')), data)
            self.assertEqual(b32decode(b32), data)

class ProtocolTestCase(TestCase):
    def check_hash_leaf(self, proto):

        # Test with wrong leaf_index type:
        with self.assertRaises(TypeError) as cm:
            proto.hash_leaf(0.0, None)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('leaf_index', int, float, 0.0)
        )
        with self.assertRaises(TypeError) as cm:
            proto.hash_leaf(1.5, None)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('leaf_index', int, float, 1.5)
        )

        # Test when leaf_index is too small:
        with self.assertRaises(ValueError) as cm:
            proto.hash_leaf(-1, None)
        self.assertEqual(
            str(cm.exception),
            'leaf_index: must be >= 0; got -1'
        )
        with self.assertRaises(ValueError) as cm:
            proto.hash_leaf(-17, None)
        self.assertEqual(
            str(cm.exception),
            'leaf_index: must be >= 0; got -17'
        )

        # Test when leaf_index is too big:
        with self.assertRaises(ValueError) as cm:
            proto.hash_leaf(proto.max_leaf_count, None)
        self.assertEqual(
            str(cm.exception),
            'leaf_index: must be < {}; got {}'.format(
                proto.max_leaf_count, proto.max_leaf_count
            )
        )
        with self.assertRaises(ValueError) as cm:
            proto.hash_leaf(proto.max_leaf_count + 1, None)
        self.assertEqual(
            str(cm.exception),
            'leaf_index: must be < {}; got {}'.format(
                proto.max_leaf_count, proto.max_leaf_count + 1
            )
        )

        # Test with wrong leaf_data type
        leaf_data = 'a' * 31415
        with self.assertRaises(TypeError) as cm:
            proto.hash_leaf(2, leaf_data)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('leaf_data', bytes, str, leaf_data)
        )

        # Test with wrong leaf_data length
        toosmall = b''
        with self.assertRaises(ValueError) as cm:
            proto.hash_leaf(2, toosmall)
        self.assertEqual(
            str(cm.exception),
            'Need 1 <= len(leaf_data) <= {}; got 0'.format(proto.leaf_size)
        )
        toobig = b'a' * (proto.leaf_size + 1)
        with self.assertRaises(ValueError) as cm:
            proto.hash_leaf(2, toobig)
        self.assertEqual(
            str(cm.exception),
            'Need 1 <= len(leaf_data) <= {}; got {}'.format(
                proto.leaf_size, proto.leaf_size + 1
            )
        )

        # Smallest possible leaf, index=0
        accum = set()
        leaf_data = b'D'
        digest = proto.hash_leaf(0, leaf_data)
        accum.add(digest)
        self.assertEqual(proto._hash_leaf(0, leaf_data, b''), digest)

        # Smallest possible leaf, index=1
        digest = proto.hash_leaf(1, leaf_data)
        self.assertNotIn(digest, accum)
        accum.add(digest)
        self.assertEqual(proto._hash_leaf(1, leaf_data, b''), digest)

        # Largest possible leaf, index=0
        leaf_data = b'D' * proto.leaf_size
        digest = proto.hash_leaf(0, leaf_data)
        self.assertNotIn(digest, accum)
        accum.add(digest)
        self.assertEqual(proto._hash_leaf(0, leaf_data, b''), digest)

        # Largest possible leaf, index=1
        digest = proto.hash_leaf(1, leaf_data)
        self.assertNotIn(digest, accum)
        accum.add(digest)
        self.assertEqual(proto._hash_leaf(1, leaf_data, b''), digest)

        # Test with challenge, smallest leaf, index=0:
        challenge = os.urandom(16)
        leaf_data = b'D'
        digest = proto.hash_leaf(0, leaf_data, challenge)
        self.assertNotIn(digest, accum)
        accum.add(digest)
        self.assertEqual(proto._hash_leaf(0, leaf_data, challenge), digest)

        # Test with challenge, smallest leaf, index=1:
        digest = proto.hash_leaf(1, leaf_data, challenge)
        self.assertNotIn(digest, accum)
        accum.add(digest)
        self.assertEqual(proto._hash_leaf(1, leaf_data, challenge), digest)

        # Test with challenge, largest leaf, index=0:
        leaf_data = b'D' * proto.leaf_size
        digest = proto.hash_leaf(0, leaf_data, challenge)
        self.assertNotIn(digest, accum)
        accum.add(digest)
        self.assertEqual(proto._hash_leaf(0, leaf_data, challenge), digest)

        # Test with challenge, largest leaf, index=1:
        digest = proto.hash_leaf(1, leaf_data, challenge)
        self.assertNotIn(digest, accum)
        accum.add(digest)
        self.assertEqual(proto._hash_leaf(1, leaf_data, challenge), digest)

        # Make sure we didn't goof
        self.assertEqual(len(accum), 8)

    def check_hash_root(self, proto):
        # Test with wrong file_size type:
        with self.assertRaises(TypeError) as cm:
            proto.hash_root(1.0, None)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('file_size', int, float, 1.0)
        )
        with self.assertRaises(TypeError) as cm:
            proto.hash_root(1.5, None)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('file_size', int, float, 1.5)
        )

        # Test when file_size is too small:
        with self.assertRaises(ValueError) as cm:
            proto.hash_root(0, None)
        self.assertEqual(
            str(cm.exception),
            'file_size: must be >= 1; got 0'
        )
        with self.assertRaises(ValueError) as cm:
            proto.hash_root(-1, None)
        self.assertEqual(
            str(cm.exception),
            'file_size: must be >= 1; got -1'
        )

        # Test when file_size is too big:
        with self.assertRaises(ValueError) as cm:
            proto.hash_root(2**53 + 1, None)
        self.assertEqual(
            str(cm.exception),
            'file_size: must be <= 9007199254740992; got 9007199254740993'
        )
        with self.assertRaises(ValueError) as cm:
            proto.hash_root(2**64 - 1, None)
        self.assertEqual(
            str(cm.exception),
            'file_size: must be <= 9007199254740992; got 18446744073709551615'
        )

        # Test with wrong leaf_hashes type:
        leaf_hashes = 'A' * proto.digest_bytes
        with self.assertRaises(TypeError) as cm:
            proto.hash_root(17, leaf_hashes)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('leaf_hashes', bytes, str, leaf_hashes)
        )

        # Test with empty leaf_hashes
        with self.assertRaises(ValueError) as cm:
            proto.hash_root(17, b'')
        self.assertEqual(str(cm.exception), 'leaf_hashes cannot be empty')

        # Test when len(leaf_hashes) is not a multiple of Protocol.digest_bytes
        with self.assertRaises(ValueError) as cm:
            proto.hash_root(17, b'A' * (proto.digest_bytes + 1))
        self.assertEqual(
            str(cm.exception),
            'len(leaf_hashes) is {}, not multiple of {}'.format(
                proto.digest_bytes + 1, proto.digest_bytes
            )
        )

        # Test low and high bounds
        with self.assertRaises(ValueError) as cm:
            proto.hash_root(proto.leaf_size + 1, b'a' * proto.digest_bytes)
        self.assertEqual(
            str(cm.exception),
            'Need 1 <= file_size <= {}; got {}'.format(
                proto.leaf_size, proto.leaf_size + 1
            )
        )
        with self.assertRaises(ValueError) as cm:
            proto.hash_root(proto.leaf_size, b'ab' * proto.digest_bytes)
        self.assertEqual(
            str(cm.exception),
            'Need {} <= file_size <= {}; got {}'.format(
                proto.leaf_size + 1,
                2 * proto.leaf_size,
                proto.leaf_size
            )
        )
        with self.assertRaises(ValueError) as cm:
            proto.hash_root(
                2 * proto.leaf_size + 1,
                b'ab' * proto.digest_bytes
            )
        self.assertEqual(
            str(cm.exception),
            'Need {} <= file_size <= {}; got {}'.format(
                proto.leaf_size + 1,
                2 * proto.leaf_size,
                2 * proto.leaf_size + 1
            )
        )

        # Test with good file_size and leaf_hashes
        accum = set()
        leaf_hashes = b'D' * proto.digest_bytes
        _id = proto.hash_root(1, leaf_hashes)
        self.assertNotIn(_id, accum)
        accum.add(_id)
        self.assertEqual(
            proto._hash_root(1, leaf_hashes),
            b32decode(_id.encode('utf-8'))
        )

        _id = proto.hash_root(2, leaf_hashes)
        self.assertNotIn(_id, accum)
        accum.add(_id)
        self.assertEqual(
            proto._hash_root(2, leaf_hashes),
            b32decode(_id.encode('utf-8'))
        )

        leaf_hashes = b'AB' * proto.digest_bytes
        _id = proto.hash_root(proto.leaf_size + 1, leaf_hashes)
        self.assertNotIn(_id, accum)
        accum.add(_id)
        self.assertEqual(
            proto._hash_root(proto.leaf_size + 1, leaf_hashes),
            b32decode(_id.encode('utf-8'))
        )

        _id = proto.hash_root(proto.leaf_size + 2, leaf_hashes)
        self.assertNotIn(_id, accum)
        accum.add(_id)
        self.assertEqual(
            proto._hash_root(proto.leaf_size + 2, leaf_hashes),
            b32decode(_id.encode('utf-8'))
        )

        # Make sure we didn't goof
        self.assertEqual(len(accum), 4)

    def sanity_check_hash_leaf(self, proto):
        """
        Random value sanity check on Protocol._hash_leaf().
        """

        # Sanity check on our crytographic claim that the
        # leaf-hash is tied to the leaf-index:
        leaf_data = os.urandom(16)
        accum = set(
            proto._hash_leaf(i, leaf_data, b'')
            for i in range(COUNT)
        )
        self.assertEqual(len(accum), COUNT)

        # Sanity check on our crytographic claim that the
        # leaf-hash is tied to the leaf-data:
        accum = set(
            proto._hash_leaf(21, os.urandom(16), b'')
            for i in range(COUNT)
        )
        self.assertEqual(len(accum), COUNT)

        # Sanity check on our crytographic claim that the
        # leaf-hash is tied to the challenge:
        accum = set(
            proto._hash_leaf(21, leaf_data, os.urandom(16))
            for i in range(COUNT)
        )
        self.assertEqual(len(accum), COUNT)

    def sanity_check_hash_root(self, proto):
        """
        Random value sanity check on Protocol._hash_root().
        """

        # Sanity check on our crytographic claim that the
        # root-hash is tied to the file-size:
        leaf_hashes = os.urandom(proto.digest_bytes)
        accum = set(
            proto._hash_root(size, leaf_hashes)
            for size in range(1, COUNT + 1)
        )
        self.assertEqual(len(accum), COUNT)

        # Sanity check on our crytographic claim that the
        # root-hash is tied to the leaf-hashes:
        accum = set(
            proto._hash_root(314159, os.urandom(proto.digest_bytes))
            for i in range(COUNT)
        )
        self.assertEqual(len(accum), COUNT)


class TestProtocol(ProtocolTestCase):
    def test_init(self):
        # Wrong `leaf_size` type:
        with self.assertRaises(TypeError) as cm:
            protocols.Protocol(1.0 * TwoMiB, 200)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('leaf_size', int, float, 1.0 * TwoMiB)
        )

        # leaf_size < MIN_LEAF_SIZE:
        with self.assertRaises(ValueError) as cm:
            protocols.Protocol(TwoMiB - 1, 200)
        self.assertEqual(
            str(cm.exception),
            'leaf_size: must be >= 2097152; got 2097151'
        )

        # leaf_size % MIN_LEAF_SIZE != 0:
        with self.assertRaises(ValueError) as cm:
            protocols.Protocol(TwoMiB * 3 // 2, 200)
        self.assertEqual(
            str(cm.exception),
            'leaf_size: must be multiple of 2097152; got 3145728'
        )

        # Wrong `digest_bits` type:
        with self.assertRaises(TypeError) as cm:
            protocols.Protocol(TwoMiB, 200.0)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('digest_bits', int, float, 200.0)
        )

        # digest_bits < 200:
        with self.assertRaises(ValueError) as cm:
            protocols.Protocol(TwoMiB, 159)
        self.assertEqual(
            str(cm.exception),
            'digest_bits: must be >= 200; got 159'
        )

        # digest_bits % 40 != 0:
        with self.assertRaises(ValueError) as cm:
            protocols.Protocol(TwoMiB, 220)
        self.assertEqual(
            str(cm.exception),
            'digest_bits: must be multiple of 40; got 220'
        )

        # Test with the good minimum values:
        proto = protocols.Protocol(TwoMiB, 200)
        self.assertEqual(proto.leaf_size, TwoMiB)
        self.assertEqual(proto.max_leaf_count,
             protocols.MAX_FILE_SIZE // TwoMiB
        )
        self.assertEqual(proto.max_leaf_count, 2**32)
        self.assertEqual(proto.digest_bits, 200)
        self.assertEqual(proto.digest_bytes, 25)
        self.assertEqual(proto.digest_b32len, 40)

        # Test with the actual v0 protocol values:
        proto = protocols.Protocol(8 * MiB, 240)
        self.assertEqual(proto.leaf_size, 8 * MiB)
        self.assertEqual(proto.max_leaf_count, 1073741824)
        self.assertEqual(proto.max_leaf_count, 2**30)
        self.assertEqual(proto.digest_bits, 240)
        self.assertEqual(proto.digest_bytes, 30)
        self.assertEqual(proto.digest_b32len, 48)

        # Test with the actual v1 protocol values:
        proto = protocols.Protocol(8 * MiB, 280)
        self.assertEqual(proto.leaf_size, 8 * MiB)
        self.assertEqual(proto.max_leaf_count, 1073741824)
        self.assertEqual(proto.max_leaf_count, 2**30)
        self.assertEqual(proto.digest_bits, 280)
        self.assertEqual(proto.digest_bytes, 35)
        self.assertEqual(proto.digest_b32len, 56)

    def test_encode(self):
        proto = protocols.Protocol(TwoMiB, 200)

        # Wrong type:
        digest = 'A' * 25
        with self.assertRaises(TypeError) as cm:
            proto.encode(digest)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('digest', bytes, str, digest)
        )

        # Wrong length:
        digest = b'A' * 30
        with self.assertRaises(ValueError) as cm:
            proto.encode(digest)
        self.assertEqual(str(cm.exception), 'len(digest) must be 25; got 30')

        digest = os.urandom(25)
        self.assertEqual(proto.encode(digest), b32dumps(digest))

    def test_decode(self):
        proto = protocols.Protocol(TwoMiB, 200)

        # Wrong type:
        _id = b'A' * 40
        with self.assertRaises(TypeError) as cm:
            proto.decode(_id)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('_id', str, bytes, _id)
        )

        # Wrong length:
        _id = 'A' * 48
        with self.assertRaises(ValueError) as cm:
            proto.decode(_id)
        self.assertEqual(str(cm.exception), 'len(_id) must be 40; got 48')

        digest = os.urandom(25)
        _id = b32dumps(digest)
        self.assertEqual(proto.decode(_id), digest)

        # Test round-trip:
        for i in range(100):
            digest = os.urandom(25)
            self.assertEqual(proto.decode(proto.encode(digest)), digest)

    def test_hash_leaf(self):
        proto = protocols.Protocol(TwoMiB, 200)
        self.check_hash_leaf(proto)
        self.sanity_check_hash_leaf(proto)

        # Smallest possible leaf, index=0
        accum = set()
        leaf_data = b'D'
        digest = proto.hash_leaf(0, leaf_data)
        accum.add(digest)
        self.assertEqual(
            digest,
            skein512(leaf_data,
                digest_bits=200,
                pers=protocols.PERS_LEAF,
                key=b'0',
            ).digest()
        )
        self.assertEqual(proto._hash_leaf(0, leaf_data, b''), digest)

        # Smallest possible leaf, index=1
        digest = proto.hash_leaf(1, leaf_data)
        self.assertNotIn(digest, accum)
        accum.add(digest)
        self.assertEqual(
            digest,
            skein512(leaf_data,
                digest_bits=200,
                pers=protocols.PERS_LEAF,
                key=b'1',
            ).digest()
        )
        self.assertEqual(proto._hash_leaf(1, leaf_data, b''), digest)

        # Largest possible leaf, index=0
        leaf_data = b'D' * TwoMiB
        digest = proto.hash_leaf(0, leaf_data)
        self.assertNotIn(digest, accum)
        accum.add(digest)
        self.assertEqual(
            digest,
            skein512(leaf_data,
                digest_bits=200,
                pers=protocols.PERS_LEAF,
                key=b'0',
            ).digest()
        )
        self.assertEqual(proto._hash_leaf(0, leaf_data, b''), digest)

        # Largest possible leaf, index=1
        digest = proto.hash_leaf(1, leaf_data)
        self.assertNotIn(digest, accum)
        accum.add(digest)
        self.assertEqual(
            digest,
            skein512(leaf_data,
                digest_bits=200,
                pers=protocols.PERS_LEAF,
                key=b'1',
            ).digest()
        )
        self.assertEqual(proto._hash_leaf(1, leaf_data, b''), digest)

        # Test with challenge, smallest leaf, index=0:
        challenge = os.urandom(16)
        leaf_data = b'D'
        digest = proto.hash_leaf(0, leaf_data, challenge)
        self.assertNotIn(digest, accum)
        accum.add(digest)
        self.assertEqual(
            digest,
            skein512(leaf_data,
                digest_bits=200,
                pers=protocols.PERS_LEAF,
                key=b'0',
                nonce=challenge,
            ).digest()
        )
        self.assertEqual(proto._hash_leaf(0, leaf_data, challenge), digest)

        # Test with challenge, smallest leaf, index=1:
        digest = proto.hash_leaf(1, leaf_data, challenge)
        self.assertNotIn(digest, accum)
        accum.add(digest)
        self.assertEqual(
            digest,
            skein512(leaf_data,
                digest_bits=200,
                pers=protocols.PERS_LEAF,
                key=b'1',
                nonce=challenge,
            ).digest()
        )
        self.assertEqual(proto._hash_leaf(1, leaf_data, challenge), digest)

        # Test with challenge, largest leaf, index=0:
        leaf_data = b'D' * TwoMiB
        digest = proto.hash_leaf(0, leaf_data, challenge)
        self.assertNotIn(digest, accum)
        accum.add(digest)
        self.assertEqual(
            digest,
            skein512(leaf_data,
                digest_bits=200,
                pers=protocols.PERS_LEAF,
                key=b'0',
                nonce=challenge,
            ).digest()
        )
        self.assertEqual(proto._hash_leaf(0, leaf_data, challenge), digest)

        # Test with challenge, largest leaf, index=1:
        digest = proto.hash_leaf(1, leaf_data, challenge)
        self.assertNotIn(digest, accum)
        accum.add(digest)
        self.assertEqual(
            digest,
            skein512(leaf_data,
                digest_bits=200,
                pers=protocols.PERS_LEAF,
                key=b'1',
                nonce=challenge,
            ).digest()
        )
        self.assertEqual(proto._hash_leaf(1, leaf_data, challenge), digest)

        # Make sure we didn't goof
        self.assertEqual(len(accum), 8)

    def test_hash_root(self):
        proto = protocols.Protocol(TwoMiB, 200)
        self.check_hash_root(proto)
        self.sanity_check_hash_root(proto)

        accum = set()

        # Smallest file-size possible for one leaf:
        leaf_hashes = b'D' * 25
        digest = proto._hash_root(1, leaf_hashes)
        self.assertNotIn(digest, accum)
        accum.add(digest)
        self.assertEqual(
            digest,
            skein512(leaf_hashes,
                digest_bits=200,
                pers=protocols.PERS_ROOT,
                key=b'1',
            ).digest()
        )
        self.assertEqual(
            proto.hash_root(1, leaf_hashes),
            b32encode(digest).decode('utf-8')
        )

        # Largest file-size possible for one leaf:
        digest = proto._hash_root(TwoMiB, leaf_hashes)
        self.assertNotIn(digest, accum)
        accum.add(digest)
        self.assertEqual(
            digest,
            skein512(leaf_hashes,
                digest_bits=200,
                pers=protocols.PERS_ROOT,
                key=b'2097152',
            ).digest()
        )
        self.assertEqual(
            proto.hash_root(TwoMiB, leaf_hashes),
            b32encode(digest).decode('utf-8')
        )

        # Smallest file-size possible for two leaves:
        leaf_hashes = b'D' * 50
        digest = proto._hash_root(TwoMiB + 1, leaf_hashes)
        self.assertNotIn(digest, accum)
        accum.add(digest)
        self.assertEqual(
            digest,
            skein512(leaf_hashes,
                digest_bits=200,
                pers=protocols.PERS_ROOT,
                key=b'2097153',
            ).digest()
        )
        self.assertEqual(
            proto.hash_root(TwoMiB + 1, leaf_hashes),
            b32encode(digest).decode('utf-8')
        )

        # Largest file-size possible for two leaves:
        digest = proto._hash_root(2 * TwoMiB, leaf_hashes)
        self.assertNotIn(digest, accum)
        accum.add(digest)
        self.assertEqual(
            digest,
            skein512(leaf_hashes,
                digest_bits=200,
                pers=protocols.PERS_ROOT,
                key=b'4194304',
            ).digest()
        )
        self.assertEqual(
            proto.hash_root(2 * TwoMiB, leaf_hashes),
            b32encode(digest).decode('utf-8')
        )

        # Make sure we didn't goof
        self.assertEqual(len(accum), 4)


class TestVersionOne(ProtocolTestCase):

    def test_md5(self):
        # To aid debugging, md5sums of UTF-8 strings needed by test vectors
        self.assertEqual(
            md5(protocols.PERS_LEAF).hexdigest(),
            '8aee35e23a5a74147b230f12123ca82e'
        )
        self.assertEqual(
            md5(protocols.PERS_ROOT).hexdigest(),
            '0445d91d37383f5384023d49e71cc629'
        )
        self.assertEqual(
            md5(str(0).encode('utf-8')).hexdigest(),
            'cfcd208495d565ef66e7dff9f98764da'
        )
        self.assertEqual(
            md5(str(1).encode('utf-8')).hexdigest(),
            'c4ca4238a0b923820dcc509a6f75849b'
        )
        self.assertEqual(
            md5(str(8388607).encode('utf-8')).hexdigest(),
            '32433904a755e2b9eb82cf167723b34f'
        )
        self.assertEqual(
            md5(str(8388608).encode('utf-8')).hexdigest(),
            '03926fda4e223707d290ac06bb996653'
        )
        self.assertEqual(
            md5(str(8388609).encode('utf-8')).hexdigest(),
            'e9b74719ce6b80c5337148d12725db03'
        )
        self.assertEqual(
            md5(str(16777215).encode('utf-8')).hexdigest(),
            '48ac8929ffdc78a66090d179ff1237d5'
        )
        self.assertEqual(
            md5(str(16777216).encode('utf-8')).hexdigest(),
            'e3e330499348f791337e9da6b534a386'
        )

        # To aid debugging, md5sums of the test vector files:
        self.assertEqual(
            md5(A).hexdigest(),
            '7fc56270e7a70fa81a5935b72eacbe29'
        )
        self.assertEqual(
            md5(B).hexdigest(),
            'd2bad3eedb424dd352d65eafbf6c79ba'
        )
        self.assertEqual(
            md5(C).hexdigest(),
            '5dd3531303dd6764acb93e5f171a4ab8'
        )
        self.assertEqual(
            md5(C + A).hexdigest(),
            '0722f8dc36d75acb602dcee8d0427ce0'
        )
        self.assertEqual(
            md5(C + B).hexdigest(),
            '77264eb6eed7777a1ee03e2601fc9f64'
        )
        self.assertEqual(
            md5(C + C).hexdigest(),
            '1fbfabdaafff31967f9a95f3a3d3c642'
        )

    def test_hash_leaf(self):
        proto = protocols.VERSION1

        self.check_hash_leaf(proto)
        self.sanity_check_hash_leaf(proto)

        # A0
        digest = proto._hash_leaf(0, A, b'')
        self.assertEqual(digest, b32loads(A0_B32))
        self.assertEqual(
            digest,
            skein512(A,
                digest_bits=280,
                pers=protocols.PERS_LEAF,
                key=b'0',
            ).digest()
        )
        self.assertEqual(proto.hash_leaf(0, A), digest)

        # A1
        digest = proto._hash_leaf(1, A, b'')
        self.assertEqual(digest, b32loads(A1_B32))
        self.assertEqual(
            digest,
            skein512(A,
                digest_bits=280,
                pers=protocols.PERS_LEAF,
                key=b'1',
            ).digest()
        )
        self.assertEqual(proto.hash_leaf(1, A), digest)

        # B0
        digest = proto._hash_leaf(0, B, b'')
        self.assertEqual(digest, b32loads(B0_B32))
        self.assertEqual(
            digest,
            skein512(B,
                digest_bits=280,
                pers=protocols.PERS_LEAF,
                key=b'0',
            ).digest()
        )
        self.assertEqual(proto.hash_leaf(0, B), digest)

        # B1
        digest = proto._hash_leaf(1, B, b'')
        self.assertEqual(digest, b32loads(B1_B32))
        self.assertEqual(
            digest,
            skein512(B,
                digest_bits=280,
                pers=protocols.PERS_LEAF,
                key=b'1',
            ).digest()
        )
        self.assertEqual(proto.hash_leaf(1, B), digest)

        # C0
        digest = proto._hash_leaf(0, C, b'')
        self.assertEqual(digest, b32loads(C0_B32))
        self.assertEqual(
            digest,
            skein512(C,
                digest_bits=280,
                pers=protocols.PERS_LEAF,
                key=b'0',
            ).digest()
        )
        self.assertEqual(proto.hash_leaf(0, C), digest)

        # C1
        digest = proto._hash_leaf(1, C, b'')
        self.assertEqual(digest, b32loads(C1_B32))
        self.assertEqual(
            digest,
            skein512(C,
                digest_bits=280,
                pers=protocols.PERS_LEAF,
                key=b'1',
            ).digest()
        )
        self.assertEqual(proto.hash_leaf(1, C), digest)

    def test_hash_root(self):
        proto = protocols.VERSION1

        self.check_hash_root(proto)
        self.sanity_check_hash_root(proto)

        A0 = b32loads(A0_B32)
        A1 = b32loads(A1_B32)
        B0 = b32loads(B0_B32)
        B1 = b32loads(B1_B32)
        C0 = b32loads(C0_B32)
        C1 = b32loads(C1_B32)

        # A
        digest = proto._hash_root(1, A0)
        self.assertEqual(
            b32encode(digest).decode('utf-8'),
            'FWV6OJYI36C5NN5DC4GS2IGWZXFCZCGJGHK35YV62LKAG7D2Z4LO4Z2S'
        )
        self.assertEqual(
            digest,
            skein512(A0,
                digest_bits=280,
                pers=protocols.PERS_ROOT,
                key=b'1',
            ).digest()
        )
        self.assertEqual(
            proto.hash_root(1, A0),
            b32encode(digest).decode('utf-8')
        )

        # B
        digest = proto._hash_root(8388607, B0)
        self.assertEqual(
            b32encode(digest).decode('utf-8'),
            'OB756PX5V32JMKJAFKIAJ4AFSFPA2WLNIK32ELNO4FJLJPEEEN6DCAAJ'
        )
        self.assertEqual(
            digest,
            skein512(B0,
                digest_bits=280,
                pers=protocols.PERS_ROOT,
                key=b'8388607',
            ).digest()
        )
        self.assertEqual(
            proto.hash_root(8388607, B0),
            b32encode(digest).decode('utf-8')
        )

        # C
        digest = proto._hash_root(8388608, C0)
        self.assertEqual(
            b32encode(digest).decode('utf-8'),
            'QSOHXCDH64IQBOG2NM67XEC6MLZKKPGBTISWWRPMCFCJ2EKMA2SMLY46'
        )
        self.assertEqual(
            digest,
            skein512(C0,
                digest_bits=280,
                pers=protocols.PERS_ROOT,
                key=b'8388608',
            ).digest()
        )
        self.assertEqual(
            proto.hash_root(8388608, C0),
            b32encode(digest).decode('utf-8')
        )

        # CA
        digest = proto._hash_root(8388609, C0 + A1)
        self.assertEqual(
            b32encode(digest).decode('utf-8'),
            'BQ5UTB33ML2VDTCTLVXK6N4VSMGGKKKDYKG24B6DOAFJB6NRSGMB5BNO'
        )
        self.assertEqual(
            digest,
            skein512(C0 + A1,
                digest_bits=280,
                pers=protocols.PERS_ROOT,
                key=b'8388609',
            ).digest()
        )
        self.assertEqual(
            proto.hash_root(8388609, C0 + A1),
            b32encode(digest).decode('utf-8')
        )

        # CB
        digest = proto._hash_root(16777215, C0 + B1)
        self.assertEqual(
            b32encode(digest).decode('utf-8'),
            'ER3LDDZ2LHMTDLOPE5XA5GEEZ6OE45VFIFLY42GEMV4TSZ2B7GJJXAIX'
        )
        self.assertEqual(
            digest,
            skein512(C0 + B1,
                digest_bits=280,
                pers=protocols.PERS_ROOT,
                key=b'16777215',
            ).digest()
        )
        self.assertEqual(
            proto.hash_root(16777215, C0 + B1),
            b32encode(digest).decode('utf-8')
        )

        # CC
        digest = proto._hash_root(16777216, C0 + C1)
        self.assertEqual(
            b32encode(digest).decode('utf-8'),
            'R6RN5KL7UBNJWR5SK5YPUKIGAOWWFMYYOVESU5DPT34X5MEK75PXXYIX'
        )
        self.assertEqual(
            digest,
            skein512(C0 + C1,
                digest_bits=280,
                pers=protocols.PERS_ROOT,
                key=b'16777216',
            ).digest()
        )
        self.assertEqual(
            proto.hash_root(16777216, C0 + C1),
            b32encode(digest).decode('utf-8')
        )


class TestOldProtocol(TestCase):
    def test_hash_leaf_index(self):
        proto = protocols.OldProtocol(TwoMiB, 200)
        accum = set()
        key = proto._hash_leaf_index(0)
        accum.add(key)
        self.assertEqual(key,
            skein512(b'0',
                digest_bits=200,
                pers=protocols.PERS_LEAF_INDEX,
            ).digest()
        )

        key = proto._hash_leaf_index(17)
        self.assertNotIn(key, accum)
        accum.add(key)
        self.assertEqual(key,
            skein512(b'17',
                digest_bits=200,
                pers=protocols.PERS_LEAF_INDEX,
            ).digest()
        )
        self.assertEqual(len(accum), 2)

        count = 25 * 1000
        accum = set(
            proto._hash_leaf_index(i) for i in range(count)
        )
        self.assertEqual(len(accum), count)

        #############################################
        # Again, this time with different digest_bits
        proto = protocols.OldProtocol(TwoMiB, 240)
        accum = set()
        key = proto._hash_leaf_index(0)
        accum.add(key)
        self.assertEqual(key,
            skein512(b'0',
                digest_bits=240,
                pers=protocols.PERS_LEAF_INDEX,
            ).digest()
        )

        key = proto._hash_leaf_index(17)
        self.assertNotIn(key, accum)
        accum.add(key)
        self.assertEqual(key,
            skein512(b'17',
                digest_bits=240,
                pers=protocols.PERS_LEAF_INDEX,
            ).digest()
        )
        self.assertEqual(len(accum), 2)

        count = 25 * 1000
        accum = set(
            proto._hash_leaf_index(i) for i in range(count)
        )
        self.assertEqual(len(accum), count)

    def test_hash_leaf(self):
        challenge = b'secret foo bar'
        proto = protocols.OldProtocol(TwoMiB, 200)
        key0 = proto._hash_leaf_index(0)
        key17 = proto._hash_leaf_index(17)

        accum = set()

        # Min leaf size, index=0
        leaf_data = b'D'
        digest = proto._hash_leaf(0, leaf_data, b'')
        self.assertNotIn(digest, accum)
        accum.add(digest)
        self.assertEqual(digest,
            skein512(key0 + leaf_data,
                digest_bits=200,
                pers=protocols.PERS_LEAF,
            ).digest()
        )

        # Min leaf size, index=17
        digest = proto._hash_leaf(17, leaf_data, b'')
        self.assertNotIn(digest, accum)
        accum.add(digest)
        self.assertEqual(digest,
            skein512(key17 + leaf_data,
                digest_bits=200,
                pers=protocols.PERS_LEAF,
            ).digest()
        )

        # With challenge, min leaf size, index=0
        digest = proto._hash_leaf(0, leaf_data, challenge)
        self.assertNotIn(digest, accum)
        accum.add(digest)
        self.assertEqual(digest,
            skein512(key0 + leaf_data,
                digest_bits=200,
                pers=protocols.PERS_LEAF,
                nonce=challenge,
            ).digest()
        )

        # With challenge, min leaf size, index=17
        digest = proto._hash_leaf(17, leaf_data, challenge)
        self.assertNotIn(digest, accum)
        accum.add(digest)
        self.assertEqual(digest,
            skein512(key17 + leaf_data,
                digest_bits=200,
                pers=protocols.PERS_LEAF,
                nonce=challenge,
            ).digest()
        )

        # Max leaf size, index=0
        leaf_data = b'D' * TwoMiB
        digest = proto._hash_leaf(0, leaf_data, b'')
        self.assertNotIn(digest, accum)
        accum.add(digest)
        self.assertEqual(digest,
            skein512(key0 + leaf_data,
                digest_bits=200,
                pers=protocols.PERS_LEAF,
            ).digest()
        )

        # Max leaf size, index=17
        digest = proto._hash_leaf(17, leaf_data, b'')
        self.assertNotIn(digest, accum)
        accum.add(digest)
        self.assertEqual(digest,
            skein512(key17 + leaf_data,
                digest_bits=200,
                pers=protocols.PERS_LEAF,
            ).digest()
        )

        # With challenge, max leaf size, index=0
        digest = proto._hash_leaf(0, leaf_data, challenge)
        self.assertNotIn(digest, accum)
        accum.add(digest)
        self.assertEqual(digest,
            skein512(key0 + leaf_data,
                digest_bits=200,
                pers=protocols.PERS_LEAF,
                nonce=challenge,
            ).digest()
        )

        # With challenge, max leaf size, index=17
        digest = proto._hash_leaf(17, leaf_data, challenge)
        self.assertNotIn(digest, accum)
        accum.add(digest)
        self.assertEqual(digest,
            skein512(key17 + leaf_data,
                digest_bits=200,
                pers=protocols.PERS_LEAF,
                nonce=challenge,
            ).digest()
        )

        # Make sure we didn't goof:
        self.assertEqual(len(accum), 8)

        # A 25k value sanity check on our crytographic claim that the
        # leaf-hash is tied to the leaf-index:
        count = 25 * 1000
        leaf_data = os.urandom(32)
        accum = set(
            proto._hash_leaf(i, leaf_data, b'')
            for i in range(count)
        )
        self.assertEqual(len(accum), count)

        # A 25k random value sanity check on our crytographic claim that the
        # leaf-hash is tied to the leaf-data:
        accum = set(
            proto._hash_leaf(21, os.urandom(32), b'')
            for i in range(count)
        )
        self.assertEqual(len(accum), count)

        # A 25k random value sanity check on our crytographic claim that the
        # leaf-hash is tied to the challenge:
        accum = set(
            proto._hash_leaf(21, leaf_data, os.urandom(16))
            for i in range(count)
        )
        self.assertEqual(len(accum), count)

    def test_hash_file_size(self):
        proto = protocols.OldProtocol(TwoMiB, 200)
        accum = set()

        key = proto._hash_file_size(1)
        self.assertNotIn(key, accum)
        accum.add(key)
        self.assertEqual(key,
            skein512(b'1',
                digest_bits=200,
                pers=protocols.PERS_FILE_SIZE,
            ).digest()
        )

        key = proto._hash_file_size(18)
        self.assertNotIn(key, accum)
        accum.add(key)
        self.assertEqual(key,
            skein512(b'18',
                digest_bits=200,
                pers=protocols.PERS_FILE_SIZE,
            ).digest()
        )
        self.assertEqual(len(accum), 2)

        count = 25 * 1000
        accum = set(
            proto._hash_file_size(size) for
            size in range(1, count + 1)
        )
        self.assertEqual(len(accum), count)

        #############################################
        # Again, this time with different digest_bits
        proto = protocols.OldProtocol(TwoMiB, 240)
        accum = set()

        key = proto._hash_file_size(1)
        self.assertNotIn(key, accum)
        accum.add(key)
        self.assertEqual(key,
            skein512(b'1',
                digest_bits=240,
                pers=protocols.PERS_FILE_SIZE,
            ).digest()
        )

        key = proto._hash_file_size(18)
        self.assertNotIn(key, accum)
        accum.add(key)
        self.assertEqual(key,
            skein512(b'18',
                digest_bits=240,
                pers=protocols.PERS_FILE_SIZE,
            ).digest()
        )
        self.assertEqual(len(accum), 2)

        count = 25 * 1000
        accum = set(
            proto._hash_file_size(size) for
            size in range(1, count + 1)
        )
        self.assertEqual(len(accum), count)

    def test_hash_root(self):
        proto = protocols.OldProtocol(TwoMiB, 200)
        key1 = proto._hash_file_size(1)
        key18 = proto._hash_file_size(18)
        keyM = proto._hash_file_size(TwoMiB)
        keyM1 = proto._hash_file_size(TwoMiB + 1)
        keyM18 = proto._hash_file_size(TwoMiB + 18)
        key2M = proto._hash_file_size(2 * TwoMiB)

        accum = set()

        # One leaf, file_size=1
        leaf_hashes = b'D' * 25
        digest = proto._hash_root(1, leaf_hashes)
        self.assertNotIn(digest, accum)
        accum.add(digest)
        self.assertEqual(digest,
            skein512(key1 + leaf_hashes,
                digest_bits=200,
                pers=protocols.PERS_ROOT,
            ).digest()
        )
        self.assertEqual(
            proto.hash_root(1, leaf_hashes),
            b32encode(digest).decode('utf-8')
        )

        # One leaf, file_size=18
        digest = proto._hash_root(18, leaf_hashes)
        self.assertNotIn(digest, accum)
        accum.add(digest)
        self.assertEqual(digest,
            skein512(key18 + leaf_hashes,
                digest_bits=200,
                pers=protocols.PERS_ROOT,
            ).digest()
        )
        self.assertEqual(
            proto.hash_root(18, leaf_hashes),
            b32encode(digest).decode('utf-8')
        )

        # One leaf, file_size=TwoMiB
        digest = proto._hash_root(TwoMiB, leaf_hashes)
        self.assertNotIn(digest, accum)
        accum.add(digest)
        self.assertEqual(digest,
            skein512(keyM + leaf_hashes,
                digest_bits=200,
                pers=protocols.PERS_ROOT,
            ).digest()
        )
        self.assertEqual(
            proto.hash_root(TwoMiB, leaf_hashes),
            b32encode(digest).decode('utf-8')
        )

        # Two leaves, file_size=(TwoMiB + 1)
        leaf_hashes = b'D' * 50
        digest = proto._hash_root(TwoMiB + 1, leaf_hashes)
        self.assertNotIn(digest, accum)
        accum.add(digest)
        self.assertEqual(digest,
            skein512(keyM1 + leaf_hashes,
                digest_bits=200,
                pers=protocols.PERS_ROOT,
            ).digest()
        )
        self.assertEqual(
            proto.hash_root(TwoMiB + 1, leaf_hashes),
            b32encode(digest).decode('utf-8')
        )

        # Two leaves, file_size=(TwoMiB + 18)
        digest = proto._hash_root(TwoMiB + 18, leaf_hashes)
        self.assertNotIn(digest, accum)
        accum.add(digest)
        self.assertEqual(digest,
            skein512(keyM18 + leaf_hashes,
                digest_bits=200,
                pers=protocols.PERS_ROOT,
            ).digest()
        )
        self.assertEqual(
            proto.hash_root(TwoMiB + 18, leaf_hashes),
            b32encode(digest).decode('utf-8')
        )

        # Two leaves, file_size=(2 * TwoMiB)
        digest = proto._hash_root(2 * TwoMiB, leaf_hashes)
        self.assertNotIn(digest, accum)
        accum.add(digest)
        self.assertEqual(digest,
            skein512(key2M + leaf_hashes,
                digest_bits=200,
                pers=protocols.PERS_ROOT,
            ).digest()
        )
        self.assertEqual(
            proto.hash_root(2 * TwoMiB, leaf_hashes),
            b32encode(digest).decode('utf-8')
        )

        # Make sure we didn't goof:
        self.assertEqual(len(accum), 6)

        # A 25k value sanity check on our crytographic claim that the
        # root-hash is tied to the file-size:
        count = 25 * 1000
        leaf_hashes = b'D' * 25
        accum = set(
            proto._hash_root(size, leaf_hashes)
            for size in range(1, count + 1)
        )
        self.assertEqual(len(accum), count)

        # A 25k random value sanity check on our crytographic claim that the
        # root-hash is tied to the leaf-hashes:
        accum = set(
            proto._hash_root(21, os.urandom(25))
            for i in range(count)
        )
        self.assertEqual(len(accum), count)
